#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

from __future__ import division

__all__ = [ 'Expression' ]

from plugin import *
from pyutilib.component.core import *
from numvalue import *
from numvalue import ZeroConstant, native_numeric_types
from var import _VarData

from sys import getrefcount

import sys
import copy
import logging
import StringIO

logger = logging.getLogger('coopr.pyomo')


def clone_expression(exp):
    if exp.is_expression():
        return copy.copy(exp)
    else:
        return exp

def chainedInequalityErrorMessage(msg=None):
    if msg is None:
        msg = "Nonconstant relational expression used in an "\
              "unexpected Boolean context."
    buf = StringIO.StringIO()
    generate_relational_expression.chainedInequality.pprint(buf)
    # We are about to raise an exception, so it's OK to reset chainedInequality
    generate_relational_expression.chainedInequality = None
    msg += """
The inequality expression:
    %s
contains non-constant terms (variables) appearing in a Boolean context, e.g.:
    if expression <= 5:
This is generally invalid.  If you want to obtain the Boolean value of
the expression based on the current variable values, explicitly evaluate
the expression, e.g.:
    if value(expression) <= 5:
or
    if value(expression <= 5):
""" % ( buf.getvalue().strip(), )
    return msg


def identify_variables(expr, include_fixed=True):
    if expr.is_expression():
        if type(expr) is _ProductExpression:
            for arg in expr._numerator:
                for var in identify_variables(arg, include_fixed):
                    yield var
            for arg in expr._denominator:
                for var in identify_variables(arg, include_fixed):
                    yield var
        else:
            for arg in expr._args:
                for var in identify_variables(arg, include_fixed):
                    yield var
    elif include_fixed or not expr.is_fixed():
        if isinstance(expr, _VarData):
            yield expr


class Expression(NumericValue):
    """An object that defines a mathematical expression that can be evaluated"""

    __slots__ = ('_args', )

    def __init__(self, name, nargs, args):
        """Construct an expression with an operation and a set of arguments"""
        # IMPT: The following three lines are equivalent to calling the
        #       basic NumericValue constructor, i.e., as follows:
        #       NumericValue.__init__(self, name, None, None, False)
        #       That particular constructor call takes a lot of time
        #       for big models, and is unnecessary because we're not
        #       validating any values.
        self.name = name
        self.domain = None
        self.value = None

        self._args=args
        if nargs and nargs != len(args):
            raise ValueError, "%s() takes exactly %d arguments (%d given)" % \
                ( name, nargs, len(args) )

    def __getstate__(self):
        result = NumericValue.__getstate__(self)
        for i in Expression.__slots__:
            result[i] = getattr(self, i)
        return result

    def pprint(self, ostream=None, nested=True, eol_flag=True, verbose=False):
        """Print this expression"""
        if ostream is None:
            ostream = sys.stdout
        if nested:
            if self.name is None:
                print >>ostream, "Unnamed-Expression" + "(",
            else:
                print >>ostream, self.name + "(",
            first=True
            for arg in self._args:
                if first==False:
                    print >>ostream, ",",
                if arg.is_expression():
                    arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
                else:
                    print >>ostream, str(arg),
                first=False
            if eol_flag==True:
                print >>ostream, ")"
            else:
                print >>ostream, ")",

    def clone(self):
        return copy.copy(self)

    def __copy__(self):
        """Clone this object using the specified arguments"""
        raise NotImplementedError, "Derived expression (%s) failed to "\
            "implement __copy__()" % ( str(self.__class__), )

    def simplify(self, model): #pragma:nocover
        print """
WARNING: Expression.simplify() has been deprecated and removed from
     Pyomo Expressions.  Please remove references to simplify() from your
     code.
"""
        return self

    #
    # this method contrast with the is_fixed() method.
    # the is_fixed() method returns true iff the value is
    # an atomic constant.
    # this method returns true iff all composite arguments
    # in this sum expression are constant, i.e., numeric
    # constants or parametrs. the parameter values can of
    # course change over time, but at any point in time,
    # they are constant. hence, the name.
    #
    def is_constant(self):
        for arg in self._args:
            if not arg.is_constant():
                return False
        return True

    def is_fixed(self):
        for arg in self._args:
            if not arg.is_fixed():
                return False
        return True

    def is_expression(self):
        return True

    def polynomial_degree(self):
        return None

    def __nonzero__(self):
        return bool(self())

    def __call__(self, exception=True):
        """Evaluate the expression"""
        try:
            return self._apply_operation(self._evaluate_arglist(self._args))
        except (ValueError, TypeError):
            if exception:
                raise
            return None

    def _evaluate_arglist(self, arglist):
        for arg in arglist:
            try:
                yield value(arg)
            except Exception, e:
                buffer = StringIO.StringIO()
                self.pprint(buffer)
                logger.error("evaluating expression: %s\n    (expression: %s)",
                             str(e), buffer.getvalue())
                raise

    def _apply_operation(self, values):
        """Method that can be overwritten to define the operation in
        this expression"""
        raise NotImplementedError, "Derived expression (%s) failed to "\
            "implement _apply_operation()" % ( str(self.__class__), )

    def __str__(self):
        return str(self.name)


class _IntrinsicFunctionExpression(Expression):
    __slots__ = ('_operator', )

    def __init__(self, name, nargs, args, operator):
        """Construct an expression with an operation and a set of arguments"""
        Expression.__init__(self, name, nargs, args)
        self._operator = operator

    def __getstate__(self):
        result = Expression.__getstate__(self)
        for i in _IntrinsicFunctionExpression.__slots__:
            result[i] = getattr(self, i)
        return result

    def __copy__(self):
        return self.__class__( self.name,
                               None,
                               tuple(clone_expression(x) for x in self._args),
                               self._operator )

    def _apply_operation(self, values):
        return self._operator(*tuple(values))

    def polynomial_degree(self):
        if self.is_fixed():
            return 0
        return None


# Should this actually be a special class, or just an instance of
# _IntrinsicFunctionExpression (like sin, cos, etc)?
class _AbsExpression(_IntrinsicFunctionExpression):

    __slots__ = ()

    def __init__(self, args):
        _IntrinsicFunctionExpression.__init__(self, 'abs', 1, args, abs)

    def __getstate__(self):
        return _IntrinsicFunctionExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

# Should this actually be a special class, or just an instance of
# _IntrinsicFunctionExpression (like sin, cos, etc)?
class _PowExpression(_IntrinsicFunctionExpression):

    __slots__ = ()

    def __init__(self, args):
        _IntrinsicFunctionExpression.__init__(self, 'pow', 2, args, pow)

    def __getstate__(self):
        return _IntrinsicFunctionExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

    def polynomial_degree(self):
        # _PowExpression is a tricky thing.  In general, a**b is
        # nonpolynomial, however, if b == 0, it is a constant
        # expression, and if a is polynomial and b is a positive
        # integer, it is also polynomial.  While we would like to just
        # call this a non-polynomial expression, these exceptions occur
        # too frequently (and in particular, a**2)
        if self._args[1].is_fixed():
            if self._args[0].is_fixed():
                return 0
            try:
                exp = int(self._args[1])
                if exp == value(self._args[1]):
                    base = self._args[0].polynomial_degree()
                    if base is not None and exp > 0:
                        return base * exp
                    elif exp == 0:
                        return 0
            except:
                pass
        return None

    def is_fixed(self):
        if self._args[1].is_fixed():
            return self._args[0].is_fixed() or bool(self._args[1] == 0)
        return False

class _LinearExpression(Expression):

    __slots__ = ()

     # the constructor for a _LinearExpression does nothing, as there are
     # no class-specific slots. thus, one is not provided - providing one
     # simply incurs overhead. the original constructor is commented out
     # below, to provide documentation as to how arguments are munged.
#    def __init__(self, name, nargs, args):
#        """Constructor"""
#        Expression.__init__(self, name, nargs, args)

    def __getstate__(self):
        return Expression.__getstate__(self)

    def polynomial_degree(self):
        # NB: We can't use max() here because None (non-polynomial)
        # overrides a numeric value (and max() just ignores it)
        degree = 0
        for x in self._args:
            x_degree = x.polynomial_degree()
            if x_degree is None:
                return None
            if x_degree > degree:
               degree = x_degree
        return degree


class _InequalityExpression(_LinearExpression):
    """An object that defines a series of less-than or
    less-than-or-equal expressions"""

    __slots__ = ('_strict', '_cloned_from')

    def __init__(self, args, strict, cloned_from):
        """Constructor"""
        _LinearExpression.__init__(self, 'ineq', None, args)
        self._strict = strict
        self._cloned_from = cloned_from
        if len(self._args)-1 != len(self._strict):
            raise ValueError, "_InequalityExpression(): len(args)-1 != " \
                  "len(strict)(%d and %d given)" % (len(args), len(strict))

    def __getstate__(self):
        result = _LinearExpression.__getstate__(self)
        for i in _InequalityExpression.__slots__:
            result[i] = getattr(self, i)
        return result

    def __nonzero__(self):
        if generate_relational_expression.chainedInequality is not None:
            raise TypeError, chainedInequalityErrorMessage()
        if not self.is_constant():
            generate_relational_expression.chainedInequality = self
            return True

        return bool(self())

    def is_relational(self):
        return True

    def __copy__(self):
        return self.__class__( [clone_expression(x) for x in self._args],
                               copy.copy(self._strict),
                               copy.copy(self._cloned_from) )

    def _apply_operation(self, values):
        """Method that defines the less-than-or-equal operation"""
        arg1 = values.next()
        for i in xrange(len(self._strict)):
            arg2 = values.next()
            if self._strict[i]:
                if not (arg1 < arg2):
                    return False
            else:
                if not (arg1 <= arg2):
                    return False
            arg1 = arg2
        return True

    def pprint(self, ostream=None, nested=True, eol_flag=True, verbose=False):
        """Print this expression"""
        if ostream is None:
            ostream = sys.stdout
        if nested:
            first=True
            for i in xrange(len(self._strict)):
                arg = self._args[i]
                if arg.is_expression():
                    arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
                else:
                    print >>ostream, str(arg),
                if self._strict[i]:
                    print >> ostream, " < ",
                else:
                    print >> ostream, " <= ",
            arg = self._args[-1]
            if arg.is_expression():
                arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
            else:
                print >>ostream, str(arg),
            if eol_flag==True:
                print >>ostream, ""


class _EqualityExpression(_LinearExpression):
    """An object that defines a equal-to expression"""

    __slots__ = ()

    def __init__(self, args):
        """Constructor"""
        _LinearExpression.__init__(self, 'eq', 2, args)

    def __getstate__(self):
        return _LinearExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

    def __nonzero__(self):
        if generate_relational_expression.chainedInequality is not None:
            raise TypeError, chainedInequalityErrorMessage()
        return bool(self())

    def is_relational(self):
        return True

    def _apply_operation(self, values):
        """Method that defines the equal-to operation"""
        return values.next() == values.next()

    def pprint(self, ostream=None, nested=True, eol_flag=True, verbose=False):
        """Print this expression"""
        if ostream is None:
            ostream = sys.stdout
        if nested:
            first = True
            for arg in self._args:
                if not first:
                    print >> ostream, " == ",
                first = False
                if arg.is_expression():
                    arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
                else:
                    print >>ostream, str(arg),
            if eol_flag==True:
                print >>ostream, ""


# It is common to generate sums of const*var (which immediately get
# thrown away by the simplifications in generate_expression): this gives
# us a pool for re-using the temporary _ProductExpression objects.
_ProdExpression_Pool = []

class _ProductExpression(Expression):
    """An object that defines a product expression"""

    __slots__ = ('_denominator','_numerator','coef')

    def __init__(self):
        """Constructor"""
        Expression.__init__(self,'prod',None, None)
        
        # generate_expression will create these for us.  Creating them
        # here is significantly wasteful as they would just end up being
        # thrown away.
        #self._denominator = []
        #self._numerator = []
        self.coef = 1

    def __getstate__(self):
        result = Expression.__getstate__(self)
        for i in _ProductExpression.__slots__:
            result[i] = getattr(self, i)
        return result

    def is_constant(self):
        for arg in self._numerator:
            if not arg.is_constant():
                return False
        for arg in self._denominator:
            if not arg.is_constant():
                return False
        return True

    def is_fixed(self):
        for arg in self._numerator:
            if not arg.is_fixed():
                return False
        for arg in self._denominator:
            if not arg.is_fixed():
                return False
        return True

    def polynomial_degree(self):
        for x in self._denominator:
            if x.polynomial_degree() != 0:
                return None
        try:
            return sum(x.polynomial_degree() for x in self._numerator)
        except TypeError:
            return None
        except ValueError:
            return None

    def invert(self):
        tmp = self._denominator
        self._denominator = self._numerator
        self._numerator = tmp
        self.coef = 1.0/self.coef

    def pprint(self, ostream=None, nested=True, eol_flag=True, verbose=False):
        """Print this expression"""
        if ostream is None:
            ostream = sys.stdout
        if nested:
            print >>ostream, self.name + "( num=(",
            first=True
            if self.coef != 1:
                print >>ostream, str(self.coef),
                first=False
            for arg in self._numerator:
                if first==False:
                    print >>ostream, ",",
                if arg.is_expression():
                    arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
                else:
                    print >>ostream, str(arg),
                first=False
            if first is True:
                print >>ostream, 1,
            print >>ostream, ")",
            if len(self._denominator) > 0:
                print >>ostream, ", denom=(",
                first=True
                for arg in self._denominator:
                    if first==False:
                        print >>ostream, ",",
                    if arg.is_expression():
                        arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
                    else:
                        print >>ostream, str(arg),
                    first=False
                print >>ostream, ")",
            print >>ostream, ")",
            if eol_flag==True:
                print >>ostream, ""

    def __copy__(self):
        """Clone this object using the specified arguments"""
        tmp = self.__class__()
        tmp.name = self.name
        tmp._numerator = [clone_expression(x) for x in self._numerator]
        tmp._denominator = [clone_expression(x) for x in self._denominator]
        tmp.coef = self.coef
        return tmp

    def __call__(self, exception=True):
        """Evaluate the expression"""
        try:
            ans = self.coef
            for n in self._evaluate_arglist(self._numerator):
                ans *= n
            for n in self._evaluate_arglist(self._denominator):
                ans /= n
            return ans
        except (TypeError, ValueError):
            if exception:
                raise
            return None


class _IdentityExpression(_LinearExpression):
    """An object that defines a identity expression"""

    __slots__ = ()

    def __init__(self, args):
        """Constructor"""
        arg_type = type(args)
        if arg_type is list or arg_type is tuple:
            Expression.__init__(self, 'identity', 1, tuple(args))
        else:
            Expression.__init__(self, 'identity', 1, (args,))

    def __getstate__(self):
        return _LinearExpression.__getstate__(self)

    def __copy__(self):
        return self.__class__( tuple(clone_expression(x) for x in self._args) )

    def _apply_operation(self, values):
        """Method that defines the identity operation"""
        return values.next()


# It is common to generate sums of sums (which immediately get
# thrown away by the simplifications in generate_expression): this gives
# us a pool for re-using the temporary _SumExpression objects.
_SumExpression_Pool = []

class _SumExpression(_LinearExpression):
    """An object that defines a weighted summation of expressions"""

    __slots__ = ('_coef','_const')

    def __init__(self):
        """Constructor"""
        _LinearExpression.__init__(self, 'sum', None, [])

        self._const = 0
        # generate_expression will create this for us.  Creating it
        # here is significantly wasteful as they would just end up being
        # thrown away.  
        #self._coef = []
        # TODO: determine if we can do something similar with _args for
        # all _LinearExpressions

    def __getstate__(self):
        result = _LinearExpression.__getstate__(self)
        for i in _SumExpression.__slots__:
            result[i] = getattr(self, i)
        return result

    def __copy__(self):
        """Clone this object using the specified arguments"""
        tmp = self.__class__()
        tmp.name = self.name
        tmp._args = [clone_expression(x) for x in self._args]
        tmp._coef = copy.copy(self._coef)
        tmp._const = self._const
        return tmp

    def scale(self, val):
        for i in xrange(len(self._coef)):
            self._coef[i] *= val
        self._const *= val

    def negate(self):
        self.scale(-1)

    def pprint(self, ostream=None, nested=True, eol_flag=True, verbose=False):
        """Print this expression"""
        if ostream is None:
            ostream = sys.stdout
        if nested:
            print >>ostream, self.name + "(",
            first=True
            if self._const != 0:
                print >>ostream, str(self._const),
                first=False
            for i, arg in enumerate(self._args):
                if first==False:
                    print >>ostream, ",",
                if self._coef[i] != 1:
                    print >>ostream, str(self._coef[i])+" * ",
                if arg.is_expression():
                    arg.pprint(ostream=ostream, nested=nested, eol_flag=False, verbose=verbose)
                else:
                    print >>ostream, str(arg),
                first=False
            print >>ostream, ")",
            if eol_flag==True:
                print >>ostream, ""

    def _apply_operation(self, values):
        """Evaluate the expression"""
        return sum(c*values.next() for c in self._coef) + self._const

def _generate_expression__clone_if_needed(obj, target):
    if getrefcount(obj) - UNREFERENCED_EXPR_COUNT == target:
        return obj
    elif getrefcount(obj) - UNREFERENCED_EXPR_COUNT > target:
        generate_expression.clone_counter += 1
        return obj.clone()
    else: #pragma:nocover
        raise RuntimeError, "Expression entered generate_expression() " \
              "with too few references (%s<0); this is indicative of a " \
              "SERIOUS ERROR in the expression reuse detection scheme." \
              % ( getrefcount(obj) - UNREFERENCED_EXPR_COUNT, )

def _generate_expression__noCloneCheck(obj, target):
    return obj

def generate_expression_bypassCloneCheck(etype, _self, other):
    global _generate_expression__noCloneCheck
    global _generate_expression__clone_if_needed
    _generate_expression__noCloneCheck, _generate_expression__clone_if_needed=\
    _generate_expression__clone_if_needed, _generate_expression__noCloneCheck

    ans = generate_expression(etype, _self, other)

    _generate_expression__noCloneCheck, _generate_expression__clone_if_needed=\
    _generate_expression__clone_if_needed, _generate_expression__noCloneCheck
    return ans



_add = 1
_sub = 2
_mul = 3
_div = 4
_pow = 5
_neg = 6
_abs = 7
_inplace = 10
_unary = _neg

_radd =         -_add
_iadd = _inplace+_add
_rsub =         -_sub
_isub = _inplace+_sub
_rmul =         -_mul
_imul = _inplace+_mul
_rdiv =         -_div
_idiv = _inplace+_div
_rpow =         -_pow
_ipow = _inplace+_pow

_old_etype_strings = {
    'add'  :          _add,
    'radd' :         -_add,
    'iadd' : _inplace+_add,
    'sub'  :          _sub,
    'rsub' :         -_sub,
    'isub' : _inplace+_sub,
    'mul'  :          _mul,
    'rmul' :         -_mul,
    'imul' : _inplace+_mul,
    'div'  :          _div,
    'rdiv' :         -_div,
    'idiv' : _inplace+_div,
    'pow'  :          _pow,
    'rpow' :         -_pow,
    'ipow' : _inplace+_pow,
    'neg'  :          _neg,
    'abs'  :          _abs,
    }
          

def generate_expression(etype, _self, other):
    if _self.is_indexed():
        raise TypeError, "Argument for expression '%s' is an indexed "\
              "numeric value specified without an index: %s\n    Is "\
              "variable or parameter '%s' defined over an index that "\
              "you did not specify?" % (etype, _self.name, _self.name)

    self_type = _self.__class__
    
    # In-place operators should only clone `self` if someone else (other
    # than the self) holds a reference to it.  This also enforces that
    # there MUST be an external `self` reference for in-place operators.
    if etype > _inplace: #and etype < 2*_inplace:#etype[0] == 'i':
        #etype = etype[1:]
        etype -= _inplace
        if _self.is_expression():
            _self = _generate_expression__clone_if_needed(_self, 1)
    else:
        if _self.is_expression():
            _self = _generate_expression__clone_if_needed(_self, 0)
        elif _self.is_constant():
            _self = _self()
            self_type = None

    #
    # First, handle the special cases of unary operators and
    # "degenerate" binary operators (subtraction)
    #
    #if etype in _generate_expression__specialCases:
    #    if etype[-1] == 'b': # This covers sub and rsub
    #        etype = etype[:-3] + 'add'
    #        multiplier = -1
    if etype >= _unary:
        if etype == _neg:
            if self_type is _SumExpression:
                _self.negate()
                return _self
            elif self_type is _ProductExpression:
                _self.coef *= -1
                return _self
            else:
                etype = _rmul#'rmul'
                other = -1
        elif etype == _abs:
            if self_type is None:
                return abs(_self)
            else:
                return _AbsExpression([_self])
    
    if other.__class__ in native_numeric_types:
        other_type = None
    else:
        try:
            other = other.as_numeric()
        except AttributeError:
            other = as_numeric(other)
        other_type = other.__class__
        if other.is_indexed():
            raise TypeError(
                "Argument for expression '%s' is an indexed numeric "
                "value\nspecified without an index:\n\t%s\nIs this "
                "value defined over an index that you did not specify?" 
                % (etype, other.name, other.name) )
        if other.is_expression():
            other = _generate_expression__clone_if_needed(other, 0)
        elif other.is_constant():
            other = other()
            other_type = None

    #
    # Binary operators can either be "normal" or "reversed"; reverse all
    # "reversed" opertors
    #
    if etype < 0:
        #
        # This may seem obvious, but if we are performing an
        # "R"-operation (i.e. reverse operation), then simply reverse
        # self and other.  This is legitimate as we are generating a
        # completely new expression here, and the _clone_if_needed logic
        # above will make sure that we don't accidentally clobber
        # someone else's expression (fragment).
        #
        etype *= -1
        _self, other = other, _self
        self_type, other_type = other_type, self_type

    if etype == _sub:
        multiplier = -1
        etype = _add
    else:
        multiplier = 1

    #
    # Now, handle all binary operators
    #
    if etype == _add:
        #
        # self + other
        #
        if self_type is _SumExpression:
            if other_type is _ProductExpression \
                     and len(other._numerator) == 1 \
                     and not other._denominator:
                _self._args += other._numerator
                _self._coef.append(multiplier*other.coef)
                _ProdExpression_Pool.append(other)
            elif other_type is _SumExpression:
                if multiplier < 0:
                    other.negate()
                _self._args += other._args
                _self._coef += other._coef
                _self._const += other._const
                _SumExpression_Pool.append(other)
            elif other_type is None: #NumericConstant:
                _self._const += multiplier * other
                if _self._const==0 and len(_self._coef)==1 and _self._coef[0]==1:
                    _SumExpression_Pool.append(_self)
                    return _self._args.pop()
            else:
                _self._args.append(other)
                _self._coef.append(multiplier)
            return _self
        elif other_type is _SumExpression:
            if multiplier < 0:
                other.negate()
            if self_type is _ProductExpression and \
                   len(_self._numerator) == 1 and \
                   not _self._denominator:
                _self._numerator += other._args
                other._args = _self._numerator
                other._coef.insert(0,_self.coef)
                _ProdExpression_Pool.append(_self)
            elif self_type is None: #NumericConstant:
                other._const += _self
                if other._const==0 and len(other._coef)==1 and other._coef[0]==1:
                    _SumExpression_Pool.append(other)
                    return other._args.pop()
            else:
                other._args.insert(0,_self)
                other._coef.insert(0,1)
            return other
        else:
            if other_type is None:
                if self_type is None:
                    return _self + multiplier * other
                elif other == 0: #NB: multiplier doesn't matter (this is x-0)!
                    return _self
                elif _SumExpression_Pool:
                    ans = _SumExpression_Pool.pop()
                else:
                    ans = _SumExpression()
                ans._const = multiplier * other
                if self_type is _ProductExpression and \
                       len(_self._numerator) == 1 and \
                       not _self._denominator:
                    ans._coef = [ _self.coef ]
                    ans._args = _self._numerator
                    _ProdExpression_Pool.append(_self)
                else:
                    ans._coef = [ 1 ]
                    ans._args = [ _self ]
            elif self_type is None:
                if _self == 0 and multiplier == 1:
                    return other
                elif _SumExpression_Pool:
                    ans = _SumExpression_Pool.pop()
                else:
                    ans = _SumExpression()
                ans._const = _self
                if other_type is _ProductExpression and \
                       len(other._numerator) == 1 and \
                       not other._denominator:
                    ans._coef = [ multiplier * other.coef ]
                    ans._args = other._numerator
                    _ProdExpression_Pool.append(other)
                else:
                    ans._coef = [ multiplier ]
                    ans._args = [ other ]
            else:
                if _SumExpression_Pool:
                    ans = _SumExpression_Pool.pop()
                else:
                    ans = _SumExpression()
                ans._const = 0
                if self_type is _ProductExpression and \
                       len(_self._numerator) == 1 and \
                       not _self._denominator:
                    ans._coef = [ _self.coef ]
                    ans._args = _self._numerator
                    _ProdExpression_Pool.append(_self)
                else:
                    ans._coef = [ 1 ]
                    ans._args = [ _self ]
                if other_type is _ProductExpression and \
                       len(other._numerator) == 1 and \
                       not other._denominator:
                    ans._coef.append( multiplier * other.coef )
                    ans._args += other._numerator
                    _ProdExpression_Pool.append(other)
                else:
                    ans._coef.append( multiplier )
                    ans._args.append( other )
            return ans

    elif etype == _mul:#'mul':
        #
        # self * other
        #
        if self_type is _ProductExpression:
            if other_type is None: #NumericConstant:
                _self.coef *= other
            elif other_type is _ProductExpression:
                _self._numerator += other._numerator
                _self._denominator += other._denominator
                _self.coef *= other.coef
                _ProdExpression_Pool.append(other)
            else:
                _self._numerator.append(other)
            ans = _self
        elif other_type is _ProductExpression:
            if self_type is None: #NumericConstant:
                other.coef *= _self
            else:
                other._numerator.insert(0,_self)
            ans = other
        else:
            if other_type is None:
                if self_type is None:
                    return _self * other
                elif other == 1:
                    return _self
                elif _ProdExpression_Pool:
                    ans = _ProdExpression_Pool.pop()
                else:
                    ans = _ProductExpression()
                ans.coef = other
                ans._numerator = [ _self ]
                ans._denominator = []
            elif self_type is None:
                if _self == 1:
                    return other
                elif _ProdExpression_Pool:
                    ans = _ProdExpression_Pool.pop()
                else:
                    ans = _ProductExpression()
                ans.coef = _self
                ans._numerator = [ other ]
                ans._denominator = []
            else:
                if _ProdExpression_Pool:
                    ans = _ProdExpression_Pool.pop()
                else:
                    ans = _ProductExpression()
                ans.coef = 1
                ans._numerator = [ _self, other ]
                ans._denominator = []

        # Special cases for simplifying expressions
        if ans.coef == 0:
            _ProdExpression_Pool.append(ans)
            return 0 #ZeroConstant
        return ans

    elif etype == _div:#'div':
        #
        # self / other
        #
        if self_type is _ProductExpression:
            if other_type is None: #NumericConstant:
                _self.coef /= other
            elif other_type is _ProductExpression:
                _self._numerator += other._denominator
                _self._denominator += other._numerator
                _self.coef /= other.coef
                _ProdExpression_Pool.append(other)
            else:
                _self._denominator.append(other)
            return _self
        elif other_type is _ProductExpression:
            other.invert()
            if self_type is None: #NumericConstant:
                if _self == 0:
                    return 0
                else:
                    other.coef *= _self
                if len(other._denominator) == 0 and len(other._numerator) == 1\
                       and other.coef == 1:
                    _ProdExpression_Pool.append(other)
                    return other._numerator.pop()
            else:
                other._numerator.insert(0,_self)
            return other
        else:
            if other_type is None: #NumericConstant
                if self_type is None:
                    return _self / other
                elif other == 1:
                    return _self
                elif _ProdExpression_Pool:
                    ans = _ProdExpression_Pool.pop()
                else:
                    ans = _ProductExpression()
                ans.coef = 1/other
                ans._numerator = [ _self ]
                ans._denominator = []
            elif self_type is None:
                if _self == 0:
                    return 0
                elif _ProdExpression_Pool:
                    ans = _ProdExpression_Pool.pop()
                else:
                    ans = _ProductExpression()
                ans.coef = _self
                ans._numerator = []
                ans._denominator = [ other ]
            else:
                if _ProdExpression_Pool:
                    ans = _ProdExpression_Pool.pop()
                else:
                    ans = _ProductExpression()
                ans.coef = 1
                ans._numerator = [ _self ]
                ans._denominator = [ other ]
            return ans    

    elif etype == _pow:#'pow':
        #
        # self ** other
        #
        if other_type is None: #NumericConstant:
            if self_type is None: #NumericConstant:
                # Special case: constant expressions should be
                # immediately reduced
                return _self ** other
            if other == 1:
                return _self
            if other == 0:
                return 1
            other = as_numeric(other)
        elif self_type is None: #NumericConstant:
            if _self == 0:
                return 0
            _self = as_numeric(_self)
        return _PowExpression((_self, other))

    else:
        if etype in _old_etype_strings:
            return generate_expression( _old_etype_strings[etype],
                                        as_numeric(_self), other )
        else:
            raise RuntimeError, "Unknown expression type '%s'" % etype

##
## "static" variables within the generate_expression function
##

# [debugging] clone_counter is a count of the number of calls to
# expr.clone() made during expression generation.
generate_expression.clone_counter = 0

# [configuration] UNREFERENCED_EXPR_COUNT is a "magic number" that
# indicates the stack depth between "normal" modeling and
# _clone_if_needed().  If an expression enters _clone_if_needed() with
# UNREFERENCED_EXPR_COUNT references, then there are no other variables
# that hold a reference to the expression and cloning is not necessary.
# If there are more references than UNREFERENCED_EXPR_COUNT, then we
# must clone the expression before operating on it.  It should be an
# error to hit _clone_if_needed() with fewer than
# UNREFERENCED_EXPR_COUNT references.
UNREFERENCED_EXPR_COUNT = 9


def _generate_relational_expression__clone_if_needed(obj):
    count = getrefcount(obj) - UNREFERENCED_RELATIONAL_EXPR_COUNT
    if count == 0 or count == CHAINED_EXPR_COUNT:
        return obj
    elif count > 0:
        generate_relational_expression.clone_counter += 1
        return obj.clone()
    else:
        raise RuntimeError, "Expression entered " \
              "generate_relational_expression() " \
              "with too few references (%s<0); this is indicative of a " \
              "SERIOUS ERROR in the expression reuse detection scheme." \
              % ( count, )

def _generate_relational_expression__noCloneCheck(obj):
    return obj


def generate_relational_expression(etype, lhs, rhs):
    cloned_from = (id(lhs), id(rhs))
    if generate_relational_expression.chainedInequality is not None:
        prevExpr = generate_relational_expression.chainedInequality
        for arg in prevExpr._cloned_from:
            if arg == id(lhs):
                lhs = prevExpr
                prevExpr = None
                cloned_from = lhs._cloned_from + (cloned_from[1],)
                break
            if arg == id(rhs):
                rhs = prevExpr
                prevExpr = None
                cloned_from = (cloned_from[0],) + rhs._cloned_from
                break
        if prevExpr is not None:
            raise TypeError, chainedInequalityErrorMessage()
        generate_relational_expression.chainedInequality = None

    if lhs.is_expression():
        lhs = _generate_relational_expression__clone_if_needed(lhs)
    elif lhs.is_indexed():
        raise TypeError, "Argument for expression '%s' is an indexed "\
              "numeric value specified without an index: %s\n    Is "\
              "variable or parameter '%s' defined over an index that "\
              "you did not specify?" % (etype, lhs.name, lhs.name)

    if rhs.is_expression():
        rhs = _generate_relational_expression__clone_if_needed(rhs)
    elif rhs.is_indexed():
        raise TypeError, "Argument for expression '%s' is an indexed "\
              "numeric value specified without an index: %s\n    Is "\
              "variable or parameter '%s' defined over an index that "\
              "you did not specify?" % (etype, rhs.name, rhs.name)

    if etype == '==':
        if lhs.is_relational() or rhs.is_relational():
            buf = StringIO.StringIO()
            if lhs.is_relational():
                lhs.pprint(buf)
            else:
                rhs.pprint(buf)
            raise TypeError, "Cannot create an EqualityExpression where "\
                  "one of the sub-expressions is a relational expression:\n"\
                  "    " + buf.getvalue().strip()
        return _EqualityExpression((lhs,rhs))
    else:
        if etype == '<=':
            strict = False
        elif etype == '<':
            strict = True
        else:
            raise ValueError, "Unknown relational expression type '%s'" % etype
        if lhs.is_relational():
            if lhs.__class__ is _InequalityExpression:
                if rhs.is_relational():
                    raise TypeError, "Cannot create an InequalityExpression "\
                          "where both sub-expressions are also relational "\
                          "expressions (we support no more than 3 terms "\
                          "in an inequality expression)."
                if len(lhs._args) > 2:
                    raise ValueError, "Cannot create an InequalityExpression "\
                          "with more than 3 terms."
                lhs._args.append(rhs)
                lhs._strict.append(strict)
                lhs._cloned_from = cloned_from
                return lhs
            else:
                buf = StringIO.StringIO()
                lhs.pprint(buf)
                raise TypeError, "Cannot create an InequalityExpression "\
                      "where one of the sub-expressions is an equality "\
                      "expression:\n    " + buf.getvalue().strip()
        elif rhs.is_relational():
            if rhs.__class__ is _InequalityExpression:
                if len(rhs._args) > 2:
                    raise ValueError, "Cannot create an InequalityExpression "\
                          "with more than 3 terms."
                rhs._args.insert(0, lhs)
                rhs._strict.insert(0, strict)
                rhs._cloned_from = cloned_from
                return rhs
            else:
                buf = StringIO.StringIO()
                rhs.pprint(buf)
                raise TypeError, "Cannot create an InequalityExpression "\
                      "where one of the sub-expressions is an equality "\
                      "expression:\n    " + buf.getvalue().strip()
        else:
            return _InequalityExpression([lhs,rhs], [strict], cloned_from)


##
## "static" variables within the generate_relational_expression function
##

# [functionality] chainedInequality allows us to generate symbolic
# expressions of the type "a < b < c".  This provides a buffer to hold
# the first inequality so the second inequality can access it later.
generate_relational_expression.chainedInequality = None

# [debugging] clone_counter is a count of the number of calls to
# expr.clone() made during expression generation.
generate_relational_expression.clone_counter = 0

# [configuration] UNREFERENCED_EXPR_COUNT is a "magic number" that
# indicates the stack depth between "normal" modeling and
# _clone_if_needed().  If an expression enters _clone_if_needed() with
# UNREFERENCED_EXPR_COUNT references, then there are no other variables
# that hold a reference to the expression and cloning is not necessary.
# If there are more references than UNREFERENCED_EXPR_COUNT, then we
# must clone the expression before operating on it.  It should be an
# error to hit _clone_if_needed() with fewer than
# UNREFERENCED_EXPR_COUNT references.
UNREFERENCED_RELATIONAL_EXPR_COUNT = 9

# [configuration] CHAINED_EXPR_COUNT is a "magic number" that indicates
# the relative stack depth for compound inequality expressions of the
# type "a < b < c".  Since Python creates the first inequality as
# temporary, and releases it after evaluating __nonzero__(), that
# expression's refcount will be CHAINED_EXPR_COUNT less than the normal
# UNREFERENCED_EXPR_COUNT.
CHAINED_EXPR_COUNT = -5


#
# If you want to completely disable clone checking (e.g., for
# line-profiling this file), set the following to "if True".
#
if False:
    _generate_relational_expression__clone_if_needed = \
        _generate_relational_expression__noCloneCheck
    _generate_expression__clone_if_needed = \
        _generate_expression__noCloneCheck


if False: #pragma:nocover
    ExpressionRegistration('<', _InequalityExpression)
    ExpressionRegistration('lt', _InequalityExpression)
    ExpressionRegistration('>', _InequalityExpression, True)
    ExpressionRegistration('gt', _InequalityExpression, True)
    ExpressionRegistration('<=', _InequalityExpression)
    ExpressionRegistration('lte', _InequalityExpression)
    ExpressionRegistration('>=', _InequalityExpression, True)
    ExpressionRegistration('gte', _InequalityExpression, True)
    ExpressionRegistration('=', _EqualityExpression)
    ExpressionRegistration('eq', _EqualityExpression)

if False: #pragma:nocover
    ExpressionRegistration('+', _SumExpression)
    ExpressionRegistration('sum', _SumExpression)
    ExpressionRegistration('*', _ProductExpression)
    ExpressionRegistration('prod', _ProductExpression)
    #ExpressionRegistration('-', _MinusExpression)
    #ExpressionRegistration('minus', _MinusExpression)
    #ExpressionRegistration('/', _DivisionExpression)
    #ExpressionRegistration('divide', _DivisionExpression)
    #ExpressionRegistration('-', _NegateExpression)
    #ExpressionRegistration('negate', _NegateExpression)
    ExpressionRegistration('abs', _AbsExpression)
    ExpressionRegistration('pow', _PowExpression)



#
# Unit Tests for pyomo.base.misc
#

import os, re, sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+os.sep+".."+os.sep+"..")
currdir= dirname(abspath(__file__))

from coopr.pyomo import *
import coopr.pyomo.scripting.pyomo as main
from pyutilib.misc import setup_redirect, reset_redirect
from pyutilib.services import registered_executable
import pyutilib.th as unittest


def rule1(model):
    return (1,model.x+model.y[1],2)
def rule2(model,i):
    return (1,model.x+model.y[1]+i,2)

class PyomoModel(unittest.TestCase):

    def setUp(self):
        self.model = AbstractModel()

    def test_construct(self):
        model = AbstractModel()
        model.a = Set(initialize=[1,2,3])
        model.A = Param(initialize=1)
        model.B = Param(model.a)
        model.x = Var(initialize=1,within=Reals)
        model.y = Var(model.a, initialize=1,within=Reals)
        model.obj = Objective(rule=lambda model: model.x+model.y[1])
        model.obj2 = Objective(model.a,rule=lambda model, i: i+model.x+model.y[1])
        model.con = Constraint(rule=rule1)
        model.con2 = Constraint(model.a, rule=rule2)
        instance = model.create()
        expr = instance.x + 1
        #instance.reset()
        OUTPUT = open(currdir+"/display.out","w")
        display(instance,ostream=OUTPUT)
        display(instance.obj,ostream=OUTPUT)
        display(instance.x,ostream=OUTPUT)
        display(instance.con,ostream=OUTPUT)
        expr.pprint(ostream=OUTPUT)
        model = AbstractModel()
        instance = model.create()
        display(instance,ostream=OUTPUT)
        OUTPUT.close()
        try:
            display(None)
            self.fail("test_construct - expected TypeError")
        except TypeError:
            pass
        self.assertFileEqualsBaseline(currdir+"/display.out",currdir+"/display.txt")


class PyomoBadModels ( unittest.TestCase ):

    def pyomo ( self, cmd, **kwargs):
        args = re.split('[ ]+', cmd )
        out = kwargs.get( 'file', None )
        if out is None:
            out = StringIO.StringIO()
        setup_redirect( out )
        os.chdir( currdir )
        output = main.run( args )
        reset_redirect()
        if not 'file' in kwargs:
            return OUTPUT.getvalue()
        return output

    @unittest.skipIf(registered_executable('glpsol') is None, "The 'glpsol' executable is not available")
    def test_uninstantiated_model_linear ( self ):
        """Run pyomo with "bad" model file.  Should fail gracefully, with
        a perhaps useful-to-the-user message."""
        return # ignore for now
        base = '%s/test_uninstantiated_model' % currdir
        fout, fbase = (base + '_linear.out', base + '.txt')
        self.pyomo('uninstantiated_model_linear.py', file=fout )
        self.assertFileEqualsBaseline( fout, fbase )

    @unittest.skipIf(registered_executable('cplex') is None, "The 'cplex' executable is not available")
    def test_uninstantiated_model_quadratic ( self ):
        """Run pyomo with "bad" model file.  Should fail gracefully, with
        a perhaps useful-to-the-user message."""
        return # ignore for now
        base = '%s/test_uninstantiated_model' % currdir
        fout, fbase = (base + '_quadratic.out', base + '.txt')
        self.pyomo('uninstantiated_model_quadratic.py --solver=cplex', file=fout )
        self.assertFileEqualsBaseline( fout, fbase )

if __name__ == "__main__":
    unittest.main()



#
# Get the directory where this script is defined, and where the baseline
# files are located.
#
import os
import sys
import string
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")

this_test_directory = dirname(abspath(__file__))+os.sep

benders_example_dir = dirname(dirname(dirname(dirname(dirname(abspath(__file__))))))+os.sep+"examples"+os.sep+"pyomo"+os.sep+"benders"+os.sep

coopr_bin_dir = dirname(dirname(dirname(dirname(dirname(dirname(dirname(abspath(__file__))))))))+os.sep+"bin"+os.sep

#
# Import the testing packages
#
import pyutilib.misc
import pyutilib.th as unittest
import pyutilib.subprocess
import coopr.plugins.solvers

cplex = None
cplex_available = False
try:
    cplex = coopr.plugins.solvers.CPLEX(keepFiles=True)
    cplex_available = (not cplex.executable() is None) and cplex.available(False)
except pyutilib.common.ApplicationError:
    cplex_available=False

gurobi = None
gurobi_available = False
try:
    gurobi = coopr.plugins.solvers.GUROBI(keepFiles=True)
    gurobi_available = (not gurobi.executable() is None) and gurobi.available(False)
except pyutilib.common.ApplicationError:
    gurobi_available=False

glpk = None
glpk_available = False
try:
    glpk = coopr.plugins.solvers.GLPK(keepFiles=True)
    glpk_available = (not glpk.executable() is None) and glpk.available(False)
except pyutilib.common.ApplicationError:
    glpk_available=False

@unittest.category('smoke')
class TestBenders(unittest.TestCase):

    def setUp(self):
        if os.path.exists(this_test_directory+'benders_cplex.out'):
            os.remove(this_test_directory+'benders_cplex.out')
        # IMPT: This step is key, as Python keys off the name of the module, not the location.
        #       So, different reference models in different directories won't be detected.
        #       If you don't do this, the symptom is a model that doesn't have the attributes
        #       that the data file expects.
        if "ReferenceModel" in sys.modules:
            del sys.modules["ReferenceModel"]

    @unittest.skipIf(not cplex_available, "The 'cplex' executable is not available")
    def test_benders_cplex(self):
        import subprocess
        out_file = open(this_test_directory+"benders_cplex.out",'w')
        os.chdir(benders_example_dir)
        subprocess.Popen(["lbin","python",benders_example_dir+"runbenders"],stdout=out_file).wait()
        os.chdir(this_test_directory)
        self.assertFileEqualsBaseline(this_test_directory+"benders_cplex.out", this_test_directory+"benders_cplex.baseline", tolerance=1e-7)


if __name__ == "__main__":
    unittest.main()



from coopr.pyomo import *
from pyutilib.component.core import implements, Plugin, alias
from coopr.pyomo.base import Var
from isomorphic_transformation import IsomorphicTransformation
from nonnegative_transform import *
from equality_transform import *

__all__ = ["StandardForm"]

class StandardForm(IsomorphicTransformation):
    """
    Creates a standard form Pyomo model that is equivalent to another model
    """

    alias("standard_form")

    def __init__(self, **kwds):
        kwds['name'] = "standard_form"
        super(StandardForm, self).__init__(**kwds)

    def __call__(self, model, **kwds):
        """
        Tranform a model to standard form
        """
        return self.apply(model, **kwds)

    def apply(self, model, **kwds):
        """
        Tranform a model to standard form
        """

        # Optional naming schemes to pass to EqualityTransform
        eq_kwds = {}
        eq_kwds["slack_names"] = kwds.pop("slack_names", "auxiliary_slack")
        eq_kwds["excess_names"] = kwds.pop("excess_names", "auxiliary_excess")
        eq_kwds["lb_names"] = kwds.pop("lb_names", "_lower_bound")
        eq_kwds["ub_names"] = kwds.pop("ub_names", "_upper_bound")

        # Optional naming schemes to pass to NonNegativeTransformation
        nn_kwds = {}
        nn_kwds["pos_suffix"] = kwds.pop("pos_suffix", "_plus")
        nn_kwds["neg_suffix"] = kwds.pop("neg_suffix", "_minus")

        nonneg = NonNegativeTransformation()
        equality = EqualityTransform()

        # Since NonNegativeTransform introduces new constraints
        # (that aren't equality constraints) we call it first.
        #
        # EqualityTransform introduces new variables, but they are#
        # constrainted to be nonnegative.
        sf = nonneg(model, **nn_kwds)
        sf = equality(sf, **eq_kwds)

        return sf



from coopr.pyomo import *
from pyutilib.component.core import implements, Plugin, Interface
from coopr.pyomo.base import IModelTransformation, Var

__all__ = ["IIsomophicTransformation", "IsomorphicTransormation"]

class IIsomorphicTransformation(Interface):


    def __call__(self, model, **kwds):
        """
        Apply the transformation. Returns a (model, variable_mapping) tuple.
        """

    def apply(self, model, **kwds):
        """
        Apply the transformation. Returns a (model, variable_mapping) tuple.
        """

class IsomorphicTransformation(Plugin):
    """
    This class represents transformations that create equivalent
    models. The values of original variables can be recovered after
    solving a problem created by a series of isomorphic
    transformations via a TransformTracker object.
    """

    implements(IIsomorphicTransformation)

    def __init__(self, **kwds):
        kwds["name"] = kwds.pop("name", "isomorphic_transform")
        Plugin.__init__(self, **kwds)

class IdentityTransformation(IsomorphicTransformation):
    """
    Returns a clone of the model.
    """

    def __call__(self, model, **kwds):
        m = model.clone()

        # Create the variable mapping
        varmap = {}

        for c in model.active_components():
            if isinstance(c, Var):
                pass

    def apply(self, model, **kwds):
        return self.__call__(model, **kwds)



#
# Unit Tests for Utility Functions
#

import os
import sys
from os.path import abspath, dirname
sys.path.insert(0, dirname(dirname(abspath(__file__)))+"/../..")
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
from coopr.pyomo import *
from nose.tools import nottest
import pickle

### Added for tests 5,6,7
from coopr.opt import SolverFactory
###

def obj_rule(model):
    return sum(model.x[a] + model.y[a] for a in model.A)
def constr_rule(model,a):
    return model.x[a] >= model.y[a]

class Test(pyutilib.th.TestCase):

    # tests the ability to pickle an abstract model prior to construction,
    # read it back it, and create an instance from it. validation is relatively
    # weak, in that it only tests the validity of an expression constructed
    # using the resulting model.
    def test_pickle1(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        model.obj = Objective(rule=obj_rule)
	model.constr = Constraint(model.A,rule=constr_rule)
        str = pickle.dumps(model)
        tmodel = pickle.loads(str)
        instance=tmodel.create()
        expr = dot_product(instance.x,instance.B,instance.y)
        OUTPUT=open(currdir+"test_expr1.out","w")
        expr.pprint(ostream=OUTPUT)
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir+"test_expr1.out",currdir+"test_expr1.txt")

    # same as above, but pickles the constructed AbstractModel and 
    # then operates on the unpickled instance.
    def test_pickle2(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        model.obj = Objective(rule=obj_rule)
	model.constr = Constraint(model.A,rule=constr_rule)
        tmp=model.create()
        str = pickle.dumps(tmp)
        instance = pickle.loads(str)
        expr = dot_product(instance.x,instance.B,instance.y)
        OUTPUT=open(currdir+"test_expr1.out","w")
        expr.pprint(ostream=OUTPUT)
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir+"test_expr1.out",currdir+"test_expr1.txt")

    # verifies that the use of lambda expressions as rules yields model instances
    # that are not pickle'able.
    @unittest.skipIf(sys.version_info[:2] < (2,6), "Skipping test because the sparse_dict repn is not supported")
    def test_pickle3(self):
        def rule1(model):
            return (1,model.x+model.y[1],2)
        def rule2(model, i):
            return (1,model.x+model.y[1]+i,2)

        model = AbstractModel()
        model.a = Set(initialize=[1,2,3])
        model.A = Param(initialize=1, mutable=True)
        model.B = Param(model.a, repn='sparse_dict', mutable=True)
        model.x = Var(initialize=1,within=Reals)
        model.y = Var(model.a, initialize=1,within=Reals)
        model.obj = Objective(rule=lambda model: model.x+model.y[1])
        model.obj2 = Objective(model.a,rule=lambda model,i: i+model.x+model.y[1])
        model.con = Constraint(rule=rule1)
        model.con2 = Constraint(model.a, rule=rule2)
        instance = model.create()
        try:
            str = pickle.dumps(instance)
            self.fail("Expected pickling error due to the use of lambda expressions - did not generate one!")
        except pickle.PicklingError:
            pass
        except TypeError:
            pass

    # verifies that we can print a constructed model and obtain identical results before and after 
    # pickling. introduced due to a test case by Gabe that illustrated __getstate__ of various 
    # modeling components was incorrectly and unexpectedly modifying object state.
    def test_pickle4(self):
    
        model = ConcreteModel()
        model.s = Set(initialize=[1,2])
        model.x = Var(within=NonNegativeReals)
        model.x_indexed = Var(model.s, within=NonNegativeReals)
        model.obj = Objective(expr=model.x + model.x_indexed[1] + model.x_indexed[2])
        model.con = Constraint(expr=model.x >= 1)
        model.con2 = Constraint(expr=model.x_indexed[1] + model.x_indexed[2] >= 4)

        inst = model.create()

        OUTPUT=open(currdir+"test_pickle4_baseline.out","w")
        inst.pprint(ostream=OUTPUT)
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir+"test_pickle4_baseline.out",currdir+"test_pickle4_baseline.txt")

        str = pickle.dumps(inst)

        OUTPUT=open(currdir+"test_pickle4_after.out","w")
        inst.pprint(ostream=OUTPUT)
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir+"test_pickle4_after.out",currdir+"test_pickle4_baseline.txt")

    # Test that an unpickled instance can be sent through the LP writer
    @nottest
    def has_cplex_lp():
        try:
            cplex = coopr.plugins.solvers.CPLEX(keepFiles=True)
            available = (not cplex.executable() is None) and cplex.available(False)
            return available
        except pyutilib.common.ApplicationError:
            return False

    @unittest.skipUnless(has_cplex_lp(), "Can't find cplex.")
    def test_pickle5(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1,2])
        model.x = Var(within=NonNegativeReals)
        model.x_indexed = Var(model.s, within=NonNegativeReals)
        model.obj = Objective(expr=model.x + model.x_indexed[1] + model.x_indexed[2])
        model.con = Constraint(expr=model.x >= 1)
        model.con2 = Constraint(expr=model.x_indexed[1] + model.x_indexed[2] >= 4)

        inst = model.create()
        strng = pickle.dumps(inst)
        up_inst = pickle.loads(strng)
        opt = SolverFactory('cplex')
        up_inst.preprocess()        # see ticket 4349
        results = opt.solve(up_inst)
        up_inst.load(results)
        assert(abs(up_inst.x.value - 1.0) < .001)
        assert(abs(up_inst.x_indexed[1].value - 0.0) < .001)
        assert(abs(up_inst.x_indexed[2].value - 4.0) < .001)

    # Test that an unpickled instance can be sent through the NL writer
    @nottest
    def has_cplex_nl():
        try:
            cplex = coopr.plugins.solvers.CPLEX(keepFiles=True)
            available = (not cplex.executable() is None) and cplex.available(False)
            asl = coopr.plugins.solvers.ASL(keepFiles=True, options={'solver':'cplexamp'})
            return available and (not asl.executable() is None) and asl.available(False)
        except pyutilib.common.ApplicationError:
            return False
    @unittest.skipUnless(has_cplex_nl(), "Can't find cplexamp.")
    def test_pickle6(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1,2])
        model.x = Var(within=NonNegativeReals)
        model.x_indexed = Var(model.s, within=NonNegativeReals)
        model.obj = Objective(expr=model.x + model.x_indexed[1] + model.x_indexed[2])
        model.con = Constraint(expr=model.x >= 1)
        model.con2 = Constraint(expr=model.x_indexed[1] + model.x_indexed[2] >= 4)

        inst = model.create()
        strng = pickle.dumps(inst)
        up_inst = pickle.loads(strng)
        up_inst.preprocess()        # see ticket 4349        
        opt = SolverFactory('cplexamp',solver_io='nl')
        results = opt.solve(up_inst)
        up_inst.load(results)
        assert(abs(up_inst.x.value - 1.0) < .001)
        assert(abs(up_inst.x_indexed[1].value - 0.0) < .001)
        assert(abs(up_inst.x_indexed[2].value - 4.0) < .001)

    # Test that an unpickled instance can be sent through the cplex python interface
    @nottest
    def module_available(module):
        try:
            __import__(module)
            return True
        except ImportError:
            return False
    @unittest.skipUnless(module_available('cplex'), "Can't cplex python interface.")
    def test_pickle7(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1,2])
        model.x = Var(within=NonNegativeReals)
        model.x_indexed = Var(model.s, within=NonNegativeReals)
        model.obj = Objective(expr=model.x + model.x_indexed[1] + model.x_indexed[2])
        model.con = Constraint(expr=model.x >= 1)
        model.con2 = Constraint(expr=model.x_indexed[1] + model.x_indexed[2] >= 4)

        inst = model.create()
        strng = pickle.dumps(inst)
        up_inst = pickle.loads(strng)
        opt = SolverFactory('cplex',solver_io='python')        
        results = opt.solve(up_inst)
        up_inst.load(results)
        assert(abs(up_inst.x.value - 1.0) < .001)
        assert(abs(up_inst.x_indexed[1].value - 0.0) < .001)
        assert(abs(up_inst.x_indexed[2].value - 4.0) < .001)
        


if __name__ == "__main__":
    unittest.main()



import pyinotify
import os
import sys

class Handler(pyinotify.ProcessEvent):
    def process_IN_MODIFY(self, event):
        if event.name.endswith('.tex'):
            print "Recompiling " + event.name + "..."
            os.popen('pdflatex ' + event.name).read()

wm = pyinotify.WatchManager()
notifier = pyinotify.Notifier(wm, Handler())
wm.add_watch(os.getcwd(), pyinotify.IN_MODIFY)
try:
    notifier.loop()
except pyinotify.NotifierError, err:
    print >>sys.stderr, err



#
# Imports
#
import sys
from coopr.pyomo import *

#
# Setup
#

model = AbstractModel()

model.NUTR = Set()

model.FOOD = Set()

model.cost = Param(model.FOOD, within=NonNegativeReals)

model.f_min = Param(model.FOOD, within=NonNegativeReals)

def f_max_valid (model, value, j):
    return value > model.f_min[j]
model.f_max = Param(model.FOOD, validate=f_max_valid)

model.n_min = Param(model.NUTR, within=NonNegativeReals)

def paramn_max (model, value, i):
    return value > model.n_min[i]
model.n_max = Param(model.NUTR, validate=paramn_max)

model.amt = Param(model.NUTR, model.FOOD, within=NonNegativeReals)

def Buy_bounds(model,i):
    return (model.f_min[i],model.f_max[i])
model.Buy = Var(model.FOOD, bounds=Buy_bounds, domain=Integers)

def Objective_rule(model):
    return summation(model.cost, model.Buy)
model.totalcost = Objective(rule=Objective_rule)

def Diet_rule(model, i):
    expr = 0
    for j in model.FOOD:
        expr = expr + model.amt[i,j] * model.Buy[j]
    return (model.n_min[i], expr, model.n_max[i])
model.Diet = Constraint(model.NUTR, rule=Diet_rule)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

"""
This file contains a library of functions needed to construct linear/piecewise-linear
constraints for a Pyomo model. All piecewise types except for SOS2, BIGM_SOS1,
BIGM_BIN were taken from the paper: Mixed-Integer Models for Non-separable Piecewise Linear
Optimization: Unifying framework and Extensions (Vielma, Nemhauser 2008).

TODO: Add regression tests for the following completed tasks
*) user not providing floats can be an major issue for BIGM's and M
*) Other TODO's
*) affine functions - BIGM_SOS1, BIGM_SOS2 ***** possible edge case bug

Possible Extensions
       *) Support for functions which are discontinuous in the domain/range. 
          This would require updating input structure and a check for which
          piecewise reps would support this.
       *) Consider another piecewise rep "SOS2_MANUAL" where we manually implement extra constraints to
          define an SOS2 set, this would be compatible with GLPK, http://winglpk.sourceforge.net/media/glpk-sos2_02.pdf
       *) double check that LOG and DLOG reps really do require (2^n)+1 points
       *) piecewise reps for multivariable functions
"""

__all__ = ['Piecewise']

import logging
import math
import itertools
import operator

from block import Block
from constraint import Constraint
from sos import SOSConstraint
from var import Var, _VarData, _VarArray
from set_types import PositiveReals, NonNegativeReals, Binary
from pyutilib.component.core import alias
from pyutilib.misc import flatten_tuple
from pyutilib.enum import Enum

#from linearize import _LinearData, \
#                      _characterize_function, \
#                      _isNonDecreasing, \
#                      _isNonIncreasing, \
#                      Bound,\
#                      WARNING_TOLERANCE

logger = logging.getLogger('coopr.pyomo')

PWRepn = Enum('SOS2','BIGM_BIN','BIGM_SOS1','CC','DCC','DLOG','LOG','MC','INC')

##############
############## the following code may be moved to a new module soon
##############
WARNING_TOLERANCE = 1e-8

Bound = Enum('Lower','Upper','Equal')

def _isNonDecreasing(vals):
    """
    checks that list of points is
    nondecreasing
    """
    it = iter(vals)
    it.next()
    op = operator.ge
    return all(itertools.starmap(op,itertools.izip(it,vals)))

def _isNonIncreasing(vals):
    """
    checks that list of points is
    nonincreasing
    """
    it = iter(vals)
    it.next()
    op = operator.le
    return all(itertools.starmap(op,itertools.izip(it,vals)))

def _characterize_function(tol,f_rule,model,points,*index):
    """
    Generates a list of range values and checks
    for convexity/concavity. Assumes domain points
    are sorted in increasing order.
    """
    
    # we cast to float to avoid possible integer division in cases where
    # the user supplies integer type points for return values
    values = [ float(f_rule(model,*flatten_tuple((index,x)))) for x in points ]
    slopes = [ (values[i]-values[i-1])/(points[i]-points[i-1]) for i in xrange(1,len(points)) ]

    # TODO: Warn when the slopes of two consecutive line
    #       segments are nearly equal since this is likely
    #       due to a user mistake and may cause issue with
    #       the solver.
    #       *** This is already done below but there 
    #           is probably a more correct way
    #           to send this warning through Pyomo
    if not all(itertools.starmap(lambda x1,x2: abs(x1-x2) > tol,itertools.izip(slopes,itertools.islice(slopes,1,None)))):
        print "WARNING: slopes of consecutive linear segments detected to be within "+str(tol)+" of one another"

    if _isNonDecreasing(slopes):
        # convex
        return 1,values
    elif _isNonIncreasing(slopes):
        # concave
        return -1,values
    return 0,values

class _LinearData(object):
    """
    This class defines the base class for all linearization
    and piecewise constraint generators..
    """

    def __init__(self, name, bound_type, domain_pts, range_pts):
        self.name = name
        self._bound_type = bound_type
        # ***Note: most (if not all) classes derived off of _LinearData
        #          assume the list of domain points is sorted.
        if not _isIncreasing(domain_pts):
            msg = "'%s' does not have a list of domain points "\
                  "that is strictly increasing"
            raise ValueError, msg % (self.name,)
        self._domain_pts = domain_pts
        self._range_pts = range_pts
        
        self.active = True
        self._constraints = []
        self._vars = []

    def activate(self):
        [con.activate() for con in self._constraints]
        self.active=True
    
    def deactivate(self):
        [con.deactivate() for con in self._constraints]
        self.active=False

    def _update_constraints(self,model,name,con_obj):
        constraint_name = self.name+'_'+name
        self._constraints.append(con_obj)
        setattr(model,constraint_name,con_obj)
    
    def _update_vars(self,model,name,var_obj):
        var_name = self.name+'_'+name
        self._vars.append(var_obj)
        setattr(model,var_name,var_obj)
        return var_obj

class _LinearizedBounds(_LinearData):
    """
    Called to create linearizations of a function
    """

    def __init__(self, *args, **kwds):
        self._grad_range_pts = kwds.pop("grad_range_pts",None)
        _LinearData.__init__(self,*args,**kwds)

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        dy_pts = self._grad_range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,dy_pts,bound_type]:
            raise RuntimeError, "_LinearizedBounds: construct() called during invalid state."

        tangent_indices = range(len(x_pts))

        def con_rule(model,i):
            LHS = y_var
            F_AT_XO = y_pts[i]
            dF_AT_XO = dy_pts[i]
            X_MINUS_XO = x_var-x_pts[i]
            if bound_type == Bound.Upper:
                return LHS <= F_AT_XO + dF_AT_XO*X_MINUS_XO
            elif bound_type == Bound.Lower:
                return LHS >= F_AT_XO + dF_AT_XO*X_MINUS_XO
            else:
                raise ValueError, "Invalid Bound for _LinearizedBounds object"
        self._update_constraints(model,'tangent_constraint', Constraint(tangent_indices,rule=con_rule))
##############
##############
##############

def _isPowerOfTwo(x):
    """
    checks that a number is a nonzero and positive power of 2
    """
    if (x <= 0):
        return false
    else:
        return ( (x & (x - 1)) == 0 )

def _isIncreasing(vals):
    """
    checks that a list of points is
    strictly increasing
    """
    it = iter(vals)
    it.next()
    op = operator.gt
    return all(itertools.starmap(op,itertools.izip(it,vals)))

def _GrayCode(nbits):
    """
    Generates a GrayCode of nbits represented
    by a list of lists
    """
    bitset = [0 for i in xrange(nbits)]
    # important that we copy bitset each time
    graycode = [list(bitset)]

    for i in xrange(2,(1<<nbits)+1):
        if i%2:
            for j in xrange(-1,-nbits,-1):
                if bitset[j]:
                    bitset[j-1]=bitset[j-1]^1
                    break
        else:
            bitset[-1]=bitset[-1]^1
        # important that we copy bitset each time
        graycode.append(list(bitset))

    return graycode

class _SimpleSinglePiecewise(_LinearData):
    """
    Called when the piecwise points list has only two points
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_SimpleSinglePiecewise: construct() called during invalid state."
        
        # create a single linear constraint
        LHS = y_var
        F_AT_XO = y_pts[0]
        dF_AT_XO = (y_pts[1]-y_pts[0])/float(x_pts[1]-x_pts[0]) # call float() to protect against integer division
        X_MINUS_XO = x_var-x_pts[0]
        if bound_type == Bound.Upper:
            expr= LHS <= F_AT_XO + dF_AT_XO*X_MINUS_XO        
        elif bound_type == Bound.Lower:
            expr= LHS >= F_AT_XO + dF_AT_XO*X_MINUS_XO        
        elif bound_type == Bound.Equal:
            expr= LHS == F_AT_XO + dF_AT_XO*X_MINUS_XO
        else:
            raise ValueError, "Invalid Bound for _SimpleSinglePiecewise object"
        self._update_constraints(model,'single_line_constraint',Constraint(expr=expr))

        # In order to enforce the same behavior as actual piecewise constraints,
        # we constrain the domain variable between the outer domain pts. But in order to
        # prevent filling the model with unecessary constraints, we only do this
        # when absolutely necessary.
        if x_var.lb < x_pts[0]:
            self._update_constraints(model,'simplified_piecewise_domain_constraint_lower',Constraint(expr=x_pts[0] <= x_var))
        if x_var.ub > x_pts[1]:
            self._update_constraints(model,'simplified_piecewise_domain_constraint_upper',Constraint(expr=x_var <= x_pts[-1]))
        
class _SimplifiedPiecewise(_LinearData):
    """
    Called when piecewise constraints are simplified due to a lower bounding
    convex function or an upper bounding concave function
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_SimplifiedPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)
        
        # create indexers
        bounds_constraint_index = range(len_x_pts-1)
        
        # create linear constraints
        def con_rule(model,i):
            LHS = y_var
            F_AT_XO = y_pts[i]
            dF_AT_XO = (y_pts[i+1]-y_pts[i])/float(x_pts[i+1]-x_pts[i]) # call float() to protect against integer division
            X_MINUS_XO = x_var-x_pts[i]
            if bound_type == Bound.Upper:
                return LHS <= F_AT_XO + dF_AT_XO*X_MINUS_XO        
            elif bound_type == Bound.Lower:
                return LHS >= F_AT_XO + dF_AT_XO*X_MINUS_XO 
            else:
                raise ValueError, "Invalid Bound for _SimplifiedPiecewise object"

        self._update_constraints(model,'simplified_piecewise_constraint',Constraint(bounds_constraint_index,rule=con_rule))
        
        # In order to enforce the same behavior as actual piecewise constraints,
        # we constrain the domain variable between the outer domain pts. But in order to
        # prevent filling the model with unecessary constraints, we only do this
        # when absolutely necessary.
        if x_var.lb < x_pts[0]:
            self._update_constraints(model,'simplified_piecewise_domain_constraint_lower',Constraint(expr=x_pts[0] <= x_var))
        if x_var.ub > x_pts[-1]:
            self._update_constraints(model,'simplified_piecewise_domain_constraint_upper',Constraint(expr=x_var <= x_pts[-1]))
    
class _SOS2Piecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the SOS2 formulation
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_SOS2Piecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)

        # create indexers
        sos2_index = range(len_x_pts)
    
        # create vars
        sos2_y = self._update_vars(model,'SOS2_y',Var(sos2_index,within=NonNegativeReals))
        
        # create piecewise constraints
        self._update_constraints(model,'SOS2_constraint1',Constraint(expr=x_var == sum(sos2_y[i]*x_pts[i] for i in sos2_index)))
        
        LHS = y_var
        RHS = sum(sos2_y[i]*y_pts[i] for i in sos2_index)
        expr = None
        if bound_type == Bound.Upper:
            expr= LHS <= RHS
        elif bound_type == Bound.Lower:
            expr= LHS >= RHS
        elif bound_type == Bound.Equal:
            expr= LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _SOS2Piecewise object"
        self._update_constraints(model,'SOS2_constraint2',Constraint(expr=expr))

        self._update_constraints(model,'SOS2_constraint3',Constraint(expr=sum(sos2_y[j] for j in sos2_index) == 1))

        self._update_constraints(model,'SOS2_constraint4',SOSConstraint(var=sos2_y, sos=2))

class _DCCPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the DCC formulation
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_DCCPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1,len_x_pts)
        vertices = range(1,len_x_pts+1)
        def polytope_verts(p):
            return xrange(p,p+2)
        
        # create vars
        lmda = self._update_vars(model,'DCC_lambda',Var(polytopes,vertices,within=PositiveReals))
        bin_y = self._update_vars(model,'DCC_bin_y',Var(polytopes,within=Binary))        
        
        # create piecewise constraints
        self._update_constraints(model,'DCC_constraint1',Constraint(expr=x_var == sum(lmda[p,v]*x_pts[v-1] for p in polytopes for v in polytope_verts(p))))

        LHS = y_var
        RHS = sum(lmda[p,v]*y_pts[v-1] for p in polytopes for v in polytope_verts(p))
        expr = None
        if bound_type == Bound.Upper:
            expr= LHS <= RHS
        elif bound_type == Bound.Lower:
            expr= LHS >= RHS
        elif bound_type == Bound.Equal:
            expr= LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _DCCPiecewise object"
        self._update_constraints(model,'DCC_constraint2',Constraint(expr=expr))
        
        def con3_rule(model,p):
            return bin_y[p] == sum(lmda[p,v] for v in polytope_verts(p))
        self._update_constraints(model,'DCC_constraint3',Constraint(polytopes,rule=con3_rule))

        self._update_constraints(model,'DCC_constraint4',Constraint(expr=sum(bin_y[p] for p in polytopes) == 1))

class _DLOGPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the DLOG formulation
    """

    def __init__(self, *args, **kwds):
        _LinearData.__init__(self,*args,**kwds)
        if not _isPowerOfTwo(len(self._domain_pts)-1):
            msg = "'%s' does not have a list of domain points "\
                  "with length (2^n)+1"
            raise ValueError, msg % (self.name,)
    
    def _Branching_Scheme(self,L):
        """
        Branching scheme for DLOG
        """
        MAX = 2**L
        mylists1 = {}
        for i in xrange(1,L+1):
            mylists1[i] = []
            start = 1
            step = MAX/(2**i)
            while(start < MAX):
                mylists1[i].extend([j for j in xrange(start,start+step)])
                start += 2*step
                
        biglist = xrange(1,MAX+1)
        mylists2 = {}
        for i in sorted(mylists1.keys()):
            mylists2[i] = []
            for j in biglist:
                if j not in mylists1[i]:
                    mylists2[i].append(j)
            mylists2[i] = sorted(mylists2[i])
    
        return mylists1, mylists2


    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_DLOGPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)
        
        # create branching schemes
        L_i = int(math.log(len_x_pts-1,2))
        B_ZERO,B_ONE = self._Branching_Scheme(L_i)
        
        # create indexers
        polytopes = range(1,len_x_pts)
        vertices = range(1,len_x_pts+1)
        bin_y_index = range(1,L_i+1)
        def polytope_verts(p):
            return xrange(p,p+2)

        # create vars
        lmda = self._update_vars(model,'DLOG_lambda',Var(polytopes,vertices,within=PositiveReals))
        bin_y = self._update_vars(model,'DLOG_bin_y',Var(bin_y_index,within=Binary))
        
        # create piecewise constraints
        self._update_constraints(model,'DLOG_constraint1',Constraint(expr=x_var == sum(lmda[p,v]*x_pts[v-1] for p in polytopes for v in polytope_verts(p))))
        
        LHS = y_var
        RHS = sum(lmda[p,v]*y_pts[v-1] for p in polytopes for v in polytope_verts(p))
        expr = None
        if bound_type == Bound.Upper:
            expr= LHS <= RHS
        elif bound_type == Bound.Lower:
            expr= LHS >= RHS
        elif bound_type == Bound.Equal:
            expr= LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _DLOGPiecewise object"
        self._update_constraints(model,'DLOG_constraint2',Constraint(expr=expr))

        self._update_constraints(model,'DLOG_constraint3',Constraint(expr=sum(lmda[p,v] for p in polytopes for v in polytope_verts(p)) == 1))

        def con4_rule(model,l):
            return sum(lmda[p,v] for p in B_ZERO[l] for v in polytope_verts(p)) <= bin_y[l]
        self._update_constraints(model,'DLOG_constraint4',Constraint(bin_y_index,rule=con4_rule))

        def con5_rule(model,l):
            return sum(lmda[p,v] for p in B_ONE[l] for v in polytope_verts(p)) <= (1-bin_y[l])
        self._update_constraints(model,'DLOG_constraint5',Constraint(bin_y_index,rule=con5_rule))


class _CCPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the CC formulation
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_CCPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1,len_x_pts)
        vertices = range(1,len_x_pts+1)
        def vertex_polys(v):
            if v == 1:
                return [v]
            if v == len_x_pts:
                return [v-1]
            else:
                return [v-1,v]
        
        # create vars
        lmda = self._update_vars(model,'CC_lambda',Var(vertices,within=NonNegativeReals))
        bin_y = self._update_vars(model,'CC_bin_y',Var(polytopes,within=Binary))

        # create piecewise constraints
        self._update_constraints(model,'CC_constraint1',Constraint(expr=x_var == sum(lmda[v]*x_pts[v-1] for v in vertices)))

        LHS = y_var
        RHS = sum(lmda[v]*y_pts[v-1] for v in vertices)
        expr = None
        if bound_type == Bound.Upper:
            expr= LHS <= RHS
        elif bound_type == Bound.Lower:
            expr= LHS >= RHS
        elif bound_type == Bound.Equal:
            expr= LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _CCPiecewise object"
        self._update_constraints(model,'CC_constraint2',Constraint(expr=expr))
        
        self._update_constraints(model,'CC_constraint3',Constraint(expr=sum(lmda[v] for v in vertices) == 1))

        def con4_rule(model,v):
            return lmda[v] <= sum(bin_y[p] for p in vertex_polys(v))
        self._update_constraints(model,'CC_constraint4',Constraint(vertices,rule=con4_rule))
        
        self._update_constraints(model,'CC_constraint5',Constraint(expr=sum(bin_y[p] for p in polytopes) == 1))

class _LOGPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the LOG formulation
    """

    def __init__(self, *args, **kwds):
        _LinearData.__init__(self,*args,**kwds)
        if not _isPowerOfTwo(len(self._domain_pts)-1):
            msg = "'%s' does not have a list of domain points "\
                  "with length (2^n)+1"
            raise ValueError, msg % (self.name,)

    def _Branching_Scheme(self,n):
        """
        Branching scheme for LOG, requires a gray code
        """
        BIGL = 2**n
        S = range(1,n+1)
        # turn the GrayCode into a dictionary indexed
        # starting at 1
        G = dict(enumerate(_GrayCode(n),start=1))
        
        L = dict((s,[k+1 for k in xrange(BIGL+1) if ((k == 0) or (G[k][s-1] == 1)) and ((k == BIGL) or (G[k+1][s-1] == 1))]) for s in S)
        R = dict((s,[k+1 for k in xrange(BIGL+1) if ((k == 0) or (G[k][s-1] == 0)) and ((k == BIGL) or (G[k+1][s-1] == 0))]) for s in S)
        
        return S,L,R
    
    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_LOGPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)
        
        # create branching schemes
        L_i = int(math.log(len_x_pts-1,2))
        S_i,B_LEFT,B_RIGHT = self._Branching_Scheme(L_i)

        # create indexers
        polytopes = range(1,len_x_pts)
        vertices = range(1,len_x_pts+1)
        bin_y_index = S_i
        
        # create vars
        lmda = self._update_vars(model,'LOG_lambda',Var(vertices,within=NonNegativeReals))
        bin_y = self._update_vars(model,'LOG_bin_y',Var(bin_y_index,within=Binary))

        # create piecewise constraints
        self._update_constraints(model,'LOG_constraint1',Constraint(expr=x_var == sum(lmda[v]*x_pts[v-1] for v in vertices)))

        LHS = y_var
        RHS = sum(lmda[v]*y_pts[v-1] for v in vertices)
        expr = None
        if bound_type == Bound.Upper:
            expr= LHS <= RHS
        elif bound_type == Bound.Lower:
            expr= LHS >= RHS
        elif bound_type == Bound.Equal:
            expr= LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _LOGPiecewise object"
        self._update_constraints(model,'LOG_constraint2',Constraint(expr=expr))

        self._update_constraints(model,'LOG_constraint3',Constraint(expr=sum(lmda[v] for v in vertices) == 1))

        def con4_rule(model,s):
            return sum(lmda[v] for v in B_LEFT[s]) <= bin_y[s]
        self._update_constraints(model,'LOG_constraint4',Constraint(bin_y_index,rule=con4_rule))

        def con5_rule(model,s):
            return sum(lmda[v] for v in B_RIGHT[s]) <= (1-bin_y[s])
        self._update_constraints(model,'LOG_constraint5',Constraint(bin_y_index,rule=con5_rule))



class _MCPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the MC formulation
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_MCPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1,len_x_pts)

        # create constants
        SLOPE = dict((p,(y_pts[p]-y_pts[p-1])/float(x_pts[p]-x_pts[p-1])) for p in polytopes)
        INTERSEPT = dict((p,y_pts[p-1] - (SLOPE[p]*x_pts[p-1])) for p in polytopes)

        # create vars
        poly_x = self._update_vars(model,'MC_poly_x',Var(polytopes))
        bin_y = self._update_vars(model,'MC_bin_y',Var(polytopes,within=Binary))

        # create piecewise constraints
        self._update_constraints(model,'MC_constraint1',Constraint(expr=x_var == sum(poly_x[p] for p in polytopes)))

        LHS = y_var
        RHS = sum(poly_x[p]*SLOPE[p]+bin_y[p]*INTERSEPT[p] for p in polytopes)
        expr = None
        if bound_type == Bound.Upper:
            expr=  LHS <= RHS
        elif bound_type == Bound.Lower:
            expr=  LHS >= RHS
        elif bound_type == Bound.Equal:
            expr=  LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _INCPiecewise object"
        self._update_constraints(model,'MC_constraint2',Constraint(expr=expr))
        
        def con3_rule(model,p):
            return bin_y[p]*x_pts[p-1] <= poly_x[p]
        self._update_constraints(model,'MC_constraint3',Constraint(polytopes,rule=con3_rule))

        def con4_rule(model,p):
            return poly_x[p]  <= bin_y[p]*x_pts[p]
        self._update_constraints(model,'MC_constraint4',Constraint(polytopes,rule=con4_rule))

        self._update_constraints(model,'MC_constraint5',Constraint(expr=sum(bin_y[p] for p in polytopes) == 1))

    
class _INCPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the INC formulation
    """

    def construct(self,model,x_var,y_var):
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_INCPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)

        # create indexers
        polytopes = range(1,len_x_pts)
        bin_y_index = range(1,len_x_pts-1)
        
        # create vars
        delta = self._update_vars(model,'INC_delta',Var(polytopes))
        delta[1].setub(1)
        delta[len_x_pts-1].setlb(0)
        bin_y = self._update_vars(model,'INC_bin_y',Var(bin_y_index,within=Binary))

        # create piecewise constraints        
        self._update_constraints(model,'INC_constraint1',Constraint(expr=x_var == x_pts[0] + sum(delta[p]*(x_pts[p]-x_pts[p-1]) for p in polytopes)))

        LHS = y_var
        RHS = y_pts[0] + sum(delta[p]*(y_pts[p]-y_pts[p-1]) for p in polytopes)
        expr = None
        if bound_type == Bound.Upper:
            expr= LHS <= RHS
        elif bound_type == Bound.Lower:
            expr= LHS >= RHS
        elif bound_type == Bound.Equal:
            expr= LHS == RHS
        else:
            raise ValueError, "Invalid Bound for _INCPiecewise object"
        self._update_constraints(model,'INC_constraint2',Constraint(expr=expr))

        def con3_rule(model,p):
            if p != polytopes[-1]:
                return delta[p+1] <= bin_y[p]
            else:
                return Constraint.Skip
        self._update_constraints(model,'INC_constraint3',Constraint(polytopes,rule=con3_rule))

        def con4_rule(model,p):
            if p != polytopes[-1]:
                return bin_y[p] <= delta[p]
            else:
                return Constraint.Skip
        self._update_constraints(model,'INC_constraint4',Constraint(polytopes,rule=con4_rule))

class _BIGMPiecewise(_LinearData):
    """
    Called to generate Piecewise constraint using the BIGM formulation
    """

    def __init__(self,*args,**kwds):
        self._binary = kwds.pop("binary",False)
        self._sos1 = kwds.pop("sos1",False)
        if (self._binary+self._sos1) != 1:
            raise ValueError, "_BIGMPiecewise must be initialized with the binary or sos1 input flag set to True (choose one)."
        _LinearData.__init__(self,*args,**kwds)

    def construct(self,model,x_var,y_var):
        #
        # The BIGM methods currently determine tightest constant M values. This method is
        # implemented in such a way that binary/sos1 variables are not created when this M is
        # zero.
        #
        tag = ""
        x_pts = self._domain_pts
        y_pts = self._range_pts
        bound_type = self._bound_type
        if None in [x_pts,y_pts,bound_type]:
            raise RuntimeError, "_BIGMPiecewise: construct() called during invalid state."
        len_x_pts = len(x_pts)

        binary = self._binary
        sos1 = self._sos1
        if binary is True:
            tag += "bin"
        elif sos1 is True:
            tag += "sos1"
        else:
            raise ValueError, "_BIGMPiecewise: construct() called during invalid state."
        
        # generate tightest bigM values
        OPT_M = {}
        OPT_M['UB'] = {}
        OPT_M['LB'] = {}

        if bound_type in [Bound.Upper,Bound.Equal]:
            OPT_M['UB'] = self._find_M(x_pts, y_pts, Bound.Upper)
        if bound_type in [Bound.Lower,Bound.Equal]:
            OPT_M['LB'] = self._find_M(x_pts, y_pts, Bound.Lower)
        
        all_keys = set(OPT_M['UB'].keys()).union(OPT_M['LB'].keys())
        full_indices = []
        full_indices.extend(range(1,len_x_pts))
        bigm_y_index = None
        bigm_y = None
        if len(all_keys) > 0:
            bigm_y_index = all_keys
            
            def y_domain():
                if binary is True:
                    return Binary
                elif sos1 is True:
                    return NonNegativeReals
            bigm_y = self._update_vars(model,tag+'_y',Var(bigm_y_index,within=y_domain()))
        
        def con1_rule(model,i):
            if bound_type in [Bound.Upper,Bound.Equal]:
                rhs = 1.0
                if i not in OPT_M['UB'].keys():
                    rhs *= 0.0
                else:
                    rhs *= OPT_M['UB'][i]*(1-bigm_y[i])
                return y_var - (y_pts[i-1] + ((y_pts[i]-y_pts[i-1])/float(x_pts[i]-x_pts[i-1]))*(x_var-x_pts[i-1])) <= rhs
            elif bound_type == Bound.Lower:
                rhs = 1.0
                if i not in OPT_M['LB'].keys():
                    rhs *= 0.0
                else:
                    rhs *= OPT_M['LB'][i]*(1-bigm_y[i])
                return y_var - (y_pts[i-1] + ((y_pts[i]-y_pts[i-1])/float(x_pts[i]-x_pts[i-1]))*(x_var-x_pts[i-1])) >= rhs

        def con2_rule(model):
            expr = [bigm_y[i] for i in xrange(1,len_x_pts) if i in all_keys]
            if len(expr) > 0:
                return sum(expr) == 1
            else:
                return Constraint.Skip

        def conAFF_rule(model,i):
            rhs = 1.0
            if i not in OPT_M['LB'].keys():
                rhs *= 0.0
            else:
                rhs *= OPT_M['LB'][i]*(1-bigm_y[i])
            return y_var - (y_pts[i-1] + ((y_pts[i]-y_pts[i-1])/float(x_pts[i]-x_pts[i-1]))*(x_var-x_pts[i-1])) >= rhs

        self._update_constraints(model,'BIGM_constraint1',Constraint(full_indices,rule=con1_rule))
        if len(all_keys) > 0:
            self._update_constraints(model,'BIGM_constraint2',Constraint(rule=con2_rule))
        if bound_type == Bound.Equal:
            self._update_constraints(model,'BIGM_constraint3',Constraint(full_indices,rule=conAFF_rule))
        
        if len(all_keys) > 0:
            if sos1 is True:
                self._update_constraints(model,'BIGM_constraint4',SOSConstraint(var=bigm_y, sos=1))

        # In order to enforce the same behavior as the other piecewise reps
        # we constrain the domain variable between the outer domain pts. But in order to
        # prevent filling the model with unecessary constraints, we only do this
        # when absolutely necessary.
        if x_var.lb < x_pts[0]:
            self._update_constraints(model,'simplified_piecewise_domain_constraint_lower',Constraint(expr=x_pts[0] <= x_var))
        if x_var.ub > x_pts[-1]:
            self._update_constraints(model,'simplified_piecewise_domain_constraint_upper',Constraint(expr=x_var <= x_pts[-1]))

    def _M_func(self,a,Fa,b,Fb,c,Fc):
        return Fa - Fb - ((a-b) * ((Fc-Fb) / float(c-b)))

    def _find_M(self,x_pts,y_pts,bound_type):
        len_x_pts = len(x_pts)
        _self_M_func = self._M_func

        M_final = {}
        for j in xrange(1,len_x_pts):
            index = j
            if (bound_type == Bound.Lower):
                M_final[index] = min( [0.0, min([_self_M_func(x_pts[k],y_pts[k],x_pts[j-1],y_pts[j-1],x_pts[j],y_pts[j]) for k in xrange(len_x_pts)])] )
            elif (bound_type == Bound.Upper):
                M_final[index] = max( [0.0, max([_self_M_func(x_pts[k],y_pts[k],x_pts[j-1],y_pts[j-1],x_pts[j],y_pts[j]) for k in xrange(len_x_pts)])] )
            else:
                raise ValueError, "Invalid Bound passed to _find_M function"
            if M_final[index] == 0.0:
                del M_final[index]
        return M_final

class Piecewise(Block):
    """
    Adds piecewise-linear constraints to a Pyomo model for functions of the form, y = f(x).
    
    Usage:
            model.const = Piecewise(model.index,model.yvar,model.xvar,**Keywords) - for indexed variables
            model.const = Piecewise(model.yvar,model.xvar,**Keywords) - for non-indexed variables
            
            where,
                    model.index - indexing set
                    model.yvar - Pyomo Var which represents the range values of f(x)
                    model.xvar - Pyomo Var which represents the domain values of f(x)

    Keywords:
    
-pw_pts={},[],()  
          A dictionary of lists (keys are index set) or a single list (for non-indexed variables) defining
          the set of domain vertices for a continuous piecewise linear function. Default is None

-pw_repn=''
          Indicates the type of piecewise representation to use. Each representation uses its own sets of different 
          constraints and variables. This can have a major impact on performance of the MIP solve.
          Choices:
                 
                 ~ + 'SOS2' - standard representation using SOS2 variables. (Default)
                 ~ + 'BIGM_BIN' - uses BigM constraints with binary variables. Theoretically tightest M
                                  values are automatically determined.
                 ~ + 'BIGM_SOS1' - uses BigM constraints with sos1 variables. Theoretically tightest M
                                  values are automatically determined.
                 ~*+ 'DCC' - Disaggregated convex combination model
                 ~*+ 'DLOG' - Logarithmic Disaggregated convex combination model
                 ~*+ 'CC' - Convex combination model
                 ~*+ 'LOG' - Logarithmic branching convex combination
                 ~*+ 'MC' - Multiple choice model
                 ~*+ 'INC' - Incremental (delta) method

                   * Source: "Mixed-Integer Models for Non-separable Piecewise Linear Optimization: 
                              Unifying framework and Extensions" (Vielma, Nemhauser 2008)
                   ~ Refer to the optional 'force_pw' kwd.

-pw_constr_type=''
          Indicates whether piecewise function represents an upper bound, lower bound, or equality constraint.
          Choices:
                 
                   + 'UB' - y variable is bounded above by piecewise function
                   + 'LB' - y variable is bounded below by piecewise function
                   + 'EQ' - y variable is equal to the piecewise function

-f_rule=f(model,i,j,...,x)
          A callable object defining the function that linearizations/piecewise constraints will be applied to.
          First argument must be a Pyomo model. The last argument is the domain value at which the function evaluates.
          Intermediate arguments are the corresponding indices of the Piecewise component (if any).
          Examples:
                   + def f(model,j,x):
                          if (j == 2):
                             return x**2 + 1.0
                          else:
                             return x**2 + 5.0
                   + f = lambda model,x: return exp(x) + value(model.p) (p is a Pyomo Param)
                   + def f(model,x):
                         return mydictionary[x]

-force_pw=True/False
          Using the given function rule and pw_pts, a check for convexity/concavity is implemented. If (1) the function is convex and
          the piecewise constraints are lower bounds or if (2) the function is concave and the piecewise constraints are upper bounds
          the piecewise constraints will be substituted for strictly linear constraints. Setting 'force_pw=True' will force the use
          of the original piecewise constraints when one of these two cases applies.
    """
    
    alias("Piecewise", "A Generator of linear and piecewise-linear constraints.")
    
    def __init__(self, *args, **kwds):
        
        # this is temporary as part of a move to user inputs
        # using Enums rather than strings
        translate_repn = {'BIGM_SOS1':PWRepn.BIGM_SOS1,\
                          PWRepn.BIGM_SOS1:PWRepn.BIGM_SOS1,\
                          'BIGM_BIN':PWRepn.BIGM_BIN,\
                          PWRepn.BIGM_BIN:PWRepn.BIGM_BIN,\
                          'SOS2':PWRepn.SOS2,\
                          PWRepn.SOS2:PWRepn.SOS2,\
                          'CC':PWRepn.CC,\
                          PWRepn.CC:PWRepn.CC,\
                          'DCC':PWRepn.DCC,\
                          PWRepn.DCC:PWRepn.DCC,\
                          'DLOG':PWRepn.DLOG,\
                          PWRepn.DLOG:PWRepn.DLOG,\
                          'LOG':PWRepn.LOG,\
                          PWRepn.LOG:PWRepn.LOG,\
                          'MC':PWRepn.MC,\
                          PWRepn.MC:PWRepn.MC,\
                          'INC':PWRepn.INC,\
                          PWRepn.INC:PWRepn.INC,\
                          None:None}
        # this is temporary as part of a move to user inputs
        # using Enums rather than strings
        translate_bound = {'UB':Bound.Upper,\
                           Bound.Upper:Bound.Upper,\
                           'LB':Bound.Lower,\
                           Bound.Lower:Bound.Lower,\
                           'EQ':Bound.Equal,\
                           Bound.Equal:Bound.Equal,\
                           None:None}

        # TODO: Update the keyword names. I think these are more clear
        #       pw_constr_type -> bound_type
        #       pw_pts -> domain_pts. 
        #       force_pw -> simpify=False (would be default True)
        #
        #       Another thougth is to move entirely to
        #       args since there are only two optional keywords (force_pw, warning_tol).
        
        # extract all keywords used by this class
        pw_points = kwds.pop('pw_pts',None)
        # translate the user input to the enum type
        pw_rep = kwds.pop('pw_repn','SOS2')
        pw_rep = translate_repn.get(pw_rep,pw_rep)
        # translate the user input to the enum type
        bound_type = kwds.pop('pw_constr_type',None)        
        bound_type = translate_bound.get(bound_type,bound_type)
        f_rule = kwds.pop('f_rule',None)
        force_pw = kwds.pop('force_pw',False)
        warning_tol = kwds.pop('warning_tol',WARNING_TOLERANCE)

        # all but the last two args should go to Block
        try:
            # Blocks have special handling when calling
            # __setattr__ with anything derived from component.
            # However, in this particular case we need to override
            # this so that these two variables don't get re-added 
            # as new model components, therefore we directly modify
            # __dict__
            self.__dict__['_domain_var'] = args[-1]
            self.__dict__['_range_var'] = args[-2]
        except IndexError:
            msg = "Piecewise component initialized with less than two arguments"
            raise TypeError, msg
        
        args = args[:-2]
        #
        # NOTE: The 'ctype' keyword argument is not defined here. This mocks
        #       what is done in PyomoModel.py, although here it feels like
        #       somewhat of a hack. The alternative is to modify the all_blocks()
        #       method to include Piecewise type components. I'm not sure which is
        #       best at this time. Although, a consequence of the current implementation
        #       is that model.pprint() labels Piecewise blocks as simply Blocks.
        #
        #kwds.setdefault('ctype', Piecewise)
        Block.__init__(self,*args,**kwds)

        # Check that the variables args are actually Pyomo Vars
        if not(isinstance(self._domain_var,_VarData) or isinstance(self._domain_var,_VarArray)):
            msg = "Piecewise component has invalid "\
                  "argument type for domain variable, %s"
            raise TypeError, msg % (repr(self._domain_var),)
        if not(isinstance(self._range_var,_VarData) or isinstance(self._range_var,_VarArray)):
            msg = "Piecewise component has invalid "\
                  "argument type for range variable, %s"
            raise TypeError, msg % (repr(self._range_var),)
        
        # Test that the keyword values make sense
        if f_rule.__class__ is not type(lambda: None):
            msg = "Piecewise component keyword 'f_rule' must "\
                  "be a function"
            raise ValueError, msg
        if bound_type not in Bound:
            msg = "Invalid value for Piecewise component "\
                  "keyword 'pw_constr_type'" 
            raise ValueError, msg
        if warning_tol.__class__ is not float:
            msg = "Invalid type '%s' for Piecewise component "\
                  "keyword 'warning_tol', which must be of type 'float'" 
            raise TypeError, msg % (type(warning_tol),)

        self._pw_rep = pw_rep
        self._bound_type = bound_type
        self._f_rule = f_rule
        self._force_pw = force_pw
        self._warning_tol = warning_tol
        
        if self.is_indexed() is False:
            if not (isinstance(pw_points, list) or isinstance(pw_points,tuple)):
                msg = "Invalid type '%s' for Piecewise component "\
                      "keyword 'pw_pts', which must be of type "\
                      "'list' or 'tuple' for non-indexed Piecewise component" 
                raise TypeError, msg % (type(pw_points),)
            self._domain_points = {None:pw_points}
        else:
            if not isinstance(pw_points,dict):
                msg = "Invalid type '%s' for Piecewise component "\
                      "keyword 'pw_pts', which must be of type "\
                      "'dict' for indexed Piecewise component"
                raise TypeError, msg % (type(pw_points),)
            self._domain_points = pw_points
    
    # GAH: Perhaps we should add 'deactivate' functionality to Blocks
    def deactivate(self):
        """
        Deactivates all _LinearData components of this class.
        """
        [comp.deactivate() for comp in self._data.itervalues()]
        self.active = False
    
    # GAH: Perhaps we should add 'activate' functionality to Blocks
    def activate(self):
        """
        Activates all _LinearData components of this class.
        """
        [comp.activate() for comp in self._data.itervalues()]
        self.active = True
    
    def construct(self, *args, **kwds):
        """
        A quick hack to call add after data has been loaded.
        """
        generate_debug_messages = (__debug__ is True) and (logger.isEnabledFor(logging.DEBUG) is True)
        self._constructed=True

        # We need to be able to add and construct new model
        # components on the fly so we make this Block behave concretely 
        self.concrete_mode()

        # construct each index of this component
        if self.is_indexed() is False:
            if generate_debug_messages:
                logger.debug("  Constructing single Piecewise component (index=None)")
            self.add( None )
        else:
            for index in self._index:
                if generate_debug_messages:
                    logger.debug("  Constructing Piecewise index "+str(index))
                self.add( index )

    def add(self, index ):

        _self_parent = self._parent()
        _self_domain_pts_index = self._domain_points[index]
        _self_xvar = None
        _self_yvar = None
        if self.is_indexed() is False:
            # allows one to mix Var and _VarData as input to 
            # non-indexed Piecewise, index would be None in this case
            # so for Var elements Var[None] is Var, but _VarData[None] would fail
            _self_xvar = self._domain_var
            _self_yvar = self._range_var
        else:
            # The following allows one to specify a Var or _VarData
            # object even with an indexed Piecewise component.
            # The most common situation will most likely be a VarArray,
            # so we try this first.
            try:
                _self_xvar = self._domain_var[index]
            except Exception:
                _self_xvar = self._domain_var
            try:
                _self_yvar = self._range_var[index]
            except Exception:
                _self_yvar = self._range_var
        
        # We add the requirment that the domain variable used by Piecewise is
        # always bounded from above and below.
        if (_self_xvar.lb is None) or (_self_xvar.ub is None):
            msg = "Piecewise '%s[%s]' found an unbounded variable "\
                  "used for the constraint domain: '%s'. "\
                  "Piecewise component requires the domain variable have "\
                  "lower and upper bounds."
            raise ValueError, msg % (self.name, index, _self_xvar)

        # Print a warning when the feasible region created by the piecewise
        # constraints does not include the domain variables bounds
        if _self_xvar.lb < min(_self_domain_pts_index):
            # print warning
            msg = "WARNING: Piecewise '%s[%s]' feasible region does not "\
                  "include the lower bound of domain variable: %s.lb = %s < %s"
            print msg % (self.name, index, _self_xvar, _self_xvar.lb, min(_self_domain_pts_index))
        if _self_xvar.ub > max(_self_domain_pts_index):
            # print warning
            msg = "WARNING: Piecewise '%s[%s]' feasible region does not "\
                  "include the upper bound of domain variable: %s.ub = %s > %s"
            print msg % (self.name, index, _self_xvar, _self_xvar.ub, max(_self_domain_pts_index))

        # generate the list of range values using the function rule
        # check if convexity or concavity holds as well
        force_simple = False
        if self.is_indexed() is False:
            character, range_pts = _characterize_function(self._warning_tol,self._f_rule, _self_parent, _self_domain_pts_index)
        else:
            character, range_pts = _characterize_function(self._warning_tol,self._f_rule, _self_parent, _self_domain_pts_index, index)

        # Make automatic simplications to the piecewise constraints
        # for the special cases of convexity and lower bound 
        # or concavity and upper bound
        if (character == -1):
            if (self._bound_type == Bound.Upper):
                force_simple = True
        elif (character == 1):
            if (self._bound_type == Bound.Lower):
                force_simple = True
                        
        # make sure the user does not want to disable the automatic
        # simplifications above
        if self._force_pw is True:
            force_simple = False
        
        comp = None
        keywords={'name':str(index),\
                  'bound_type':self._bound_type,\
                  'domain_pts':_self_domain_pts_index,\
                  'range_pts':range_pts}
        if force_simple is True:
            # In the case where the feasible region is convex or concave (and
            # the user does not want to force the use of piecewise constraints)
            # use simple linear constraints.
            comp = _SimplifiedPiecewise(**keywords)
        else:
            if len(_self_domain_pts_index) == 2:
                # Always use a simple single line constraint when 
                # only two points are present in the piecewise list
                comp = _SimpleSinglePiecewise(**keywords)
            else:
                # generate piecewise constraints
                if self._pw_rep == PWRepn.SOS2:
                    comp = _SOS2Piecewise(**keywords)
                elif self._pw_rep == PWRepn.INC:
                    comp = _INCPiecewise(**keywords)
                elif self._pw_rep == PWRepn.MC:
                    comp = _MCPiecewise(**keywords)
                elif self._pw_rep == PWRepn.DCC:
                    comp = _DCCPiecewise(**keywords)
                elif self._pw_rep == PWRepn.DLOG:
                    comp = _DLOGPiecewise(**keywords)
                elif self._pw_rep == PWRepn.CC:
                    comp = _CCPiecewise(**keywords)
                elif self._pw_rep == PWRepn.LOG:
                    comp = _LOGPiecewise(**keywords)
                elif self._pw_rep == PWRepn.BIGM_BIN:
                    comp = _BIGMPiecewise(binary=True, **keywords)
                elif self._pw_rep == PWRepn.BIGM_SOS1:
                    comp = _BIGMPiecewise(sos1=True, **keywords)
                else:
                    msg = "Piecewise '%s[%s]' does not have a valid "\
                          "piecewise representation: '%s'"
                    raise ValueError, msg % (self.name, index, self._pw_rep)

        self._data[index] = comp
        comp.construct(self,_self_xvar,_self_yvar)

    # TODO: Define what this means
    def reset(self):
        raise NotImplementedError, "Piecewise component has not yet implemented "\
                                 "the reset method"




#
# Imports
#
import sys
sys.path.append("../../..")
from coopr.pyomo import *
##pyomo.set_debugging()

##
## Creating a model
##
model = AbstractModel()

##
## Declaring Sets
##
#
# An unordered set of arbitrary objects can be defined by creating a Set()
# object:
#
model.A = Set()
#
# An index set of sets can also be specified by providing sets as options
# to the Set() object:
# 
model.B = Set()
model.C = Set(model.A,model.B)
#
# Set declarations can also use standard set operations to declare
# a set in a constructive fashion:
#
model.D = model.A | model.B
model.E = model.B & model.A
model.F = model.A - model.B
model.G = model.A ^ model.B
#
# Also, set cross-products can be specified as A*B
#
model.H = model.A * model.B
#
# Note that this is different from the following, which specifies that Hsub
# is a subset of this cross-product.
#
model.Hsub = Set(within=model.A * model.B)

##
## Data for Simple Sets
##
#
# A set can be constructed with the _initialize_ option, which is a function 
# that accepts the set indices and model and returns the value of that set 
# element:
#
def I_init(model):
    ans=[]
    for a in model.A:
        for b in model.B:
            ans.append( (a,b) )
    return ans
model.I = model.A*model.B
model.I.initialize = I_init
#
# Note that the set model.I is not created when this set object is
# constructed.  Instead, I_init() is called during the construction of a
# problem instance.
#
# A set can also be explicitly constructed by add set elements:
#
model.J = Set()
model.J.add(1,4,9)
#
# The _initialize_ option can also be used to specify the values in
# a set.  These default values may be overriden by later construction
# steps, or by data in an input file:
#
model.K = Set(initialize=[1,4,9])
model.K_2 = Set(initialize=[(1,4),(9,16)],dimen=2)
#
# Validation of set data is supported in two different ways.  First, a
# superset can be specified with the _within_ option:
#
model.L = Set(within=model.A)
#
# Validation of set data can also be performed with the _validate_ option,
# which is a function that returns True if a data belongs in this set:
#
def M_validate(model, value):
    return value in model.A
model.M = Set(validate=M_validate)
#
# Although the _within_ option is convenient, it can force the creation of
# a temporary set.  For example, consider the declaration
#
model.N = Set(within=model.A*model.B)
#
# In this example, the cross-product of sets A and B is needed to validate
# the members of set C.  Pyomo creates this set implicitly and uses
# it for validation.  By contrast, a simple validation function could be used
# in this example, though with a less intuitive syntax:
#
def O_validate(model, value):
    return value[0] in model.A and value[1] in model.B
model.O = Set(validate=O_validate)

##
## Data for Set Arrays
##
#
# A set array can be constructed with the _initialize_ option, which is a 
# function that accepts the set indices and model and returns the set for that 
# array index:
#
def P_init(model, i, j):
    return range(0,i*j)
model.P = Set(model.B,model.B)
model.P.initialize = P_init
#
# A set array CANNOT be explicitly constructed by adding set elements
# to individual arrays.  For example, the following is invalid:
#
#   model.Q = Set(model.B)
#   model.Q[2].add(4)
#   model.Q[4].add(16)
#
# The reason is that the line
#
#   model.Q = Set(model.B)
#
# declares set Q with an abstract index set B.  However, B is not initialized
# until the 'model.create()' call is executed at the end of this file.  We
# could, however, execute 
#
#   model.Q[2].add(4)
#   model.Q[4].add(16)
#
# after the execution of 'model.create()'.
#
# The _initialize_ option can also be used to specify the values in
# a set array.  These default values are defined in a dictionary, which 
# specifies how each array element is initialized:
#
R_init={}
R_init[2] = [1,3,5]
R_init[3] = [2,4,6]
R_init[4] = [3,5,7]
model.R = Set(model.B,initialize=R_init)
#
# Validation of a set array is supported with the _within_ option.  The
# elements of all sets in the array must be in this set:
#
model.S = Set(model.B, within=model.A)
#
# Validation of set arrays can also be performed with the _validate_ option.
# This is applied to all sets in the array:
#
def T_validate(model, value):
    return value in model.A
model.T = Set(model.B, validate=M_validate)

##
## Set options
##
#
# By default, sets are unordered.  That is, the internal representation
# may place the set elements in any order.  In some cases, we need to know
# the order in which set elements are declared.  In such cases, we can declare
# a set to be ordered with an additional constructor option.  
#
# An ordered set can take a initialization function with an additional option 
# that specifies the index into the ordered set.  In this case, the function is 
# called repeatedly to construct each element in the set:
# 
def U_init(model, z):
    if z==6:
        return Set.End
    if z==1:
        return 1
    else:
        return model.U[z-1]*z
model.U = Set(ordered=True, initialize=U_init)
#
# This example can be generalized to array sets.  Note that in this case
# we can use ordered sets to to index the array, thereby guaranteeing that
# data has been filled.  The following example illustrates the use of the
# RangeSet(a,b) object, which generates an ordered set from 'a' to 'b'
# (inclusive).
#
def V_init(model, z, i):
    if z==6:
        return Set.End
    if i==1:
        if z==1:
            return 1
        else:
            return z
    return model.V[i-1][z]+z-1
model.V = Set(RangeSet(1,4), initialize=V_init, ordered=True)

##
## Process an input file and confirm that we get appropriate 
## set instances.
##
instance = model.create("set.dat")
instance.pprint()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['RangeSet']

import math
import itertools
import sets
import expr
import set_types
from misc import apply_indexed_rule
from numvalue import value
from pyutilib.component.core import *

def count(start=0, step=1):
    n = start
    while True:
        yield n
        n += step

class RangeSetValidator(object):

    def __init__(self, start, end, step):
        self.start=start
        self.end=end
        self.step=step

    def __call__(self, model, val):
        if not type(val) in [int, float, long]:
            return False
        if val + 1e-7 < self.start:
            return False
        if val > self.end+1e-7:
            return False
        if type(self.start) is int and type(self.end) is int and type(self.step) is int and (val-self.start)%self.step != 0:
            return False
        return True


class RangeSet(sets._SetContainer):
    """A set that represents a list of numeric values"""

    alias("RangeSet", "A sequence of numeric values.  RangeSet(start,end,step) is a sequence starting a value 'start', and increasing in values by 'step' until a value greater than or equal to 'end' is reached.")

    def __init__(self,*args,**kwds):
        """Construct a list of integers"""
        tmp=()
        sets._SetContainer.__init__(self,*tmp,**kwds)
        if len(args) == 0:
            raise RuntimeError, "Attempting to construct a RangeSet object with no arguments!"

        self._type=RangeSet
        if len(args) == 1:
            self._start=1
            self._end=args[0]
            self._step=1
        elif len(args) == 2:
            self._start=args[0]
            self._end=args[1]
            self._step=1
        else:
            self._start=args[0]
            self._end=args[1]
            self._step=args[2]
        self.ordered=True
        self.value = None
        self.virtual = True
        self.concrete = True
        self._len = 0

    def construct(self, values=None):
        if self._constructed:
            return
        self._constructed=True
        if isinstance(self._start,expr.Expression):
            self._start_val = self._start()
        else:
            self._start_val = value(self._start)

        if isinstance(self._end,expr.Expression):
            self._end_val = self._end()
        else:
            self._end_val = value(self._end)

        if isinstance(self._step,expr.Expression):
            self._step_val = self._step()
        else:
            self._step_val = value(self._step)

        if type(self._start_val) is int and type(self._step) is int and type(self._end_val) is int:
            self.domain = set_types.Integers
        else:
            self.domain = set_types.Reals
        lb = self._start_val

        if self.filter is None and self.validate is None:
            self._len = int(math.floor((self._end_val-self._start_val+self._step_val+1e-7)//self._step_val))
            ub = self._start_val + (self._len-1)*self._step_val
        else:
            ub = self._start_val
            ctr=0
            for i in self:
                ub = i
                ctr += 1
            self._len = ctr
        self._bounds = (lb,ub)

    def __len__(self):
        return self._len

    def __iter__(self):
        if not self._constructed:
            raise RuntimeError, "A RangeSet class must be initialized before a user can iterate over its elements."
        if self.filter is None and self.validate is None:
            #for i in itertools.islice(count(), (self._end_val-self._start_val+self._step_val+1e-7)//self._step_val):
            for i in xrange(int((self._end_val-self._start_val+self._step_val+1e-7)//self._step_val)):
                yield self._start_val + i*self._step_val
        else:
            #for i in itertools.islice(count(), (self._end_val-self._start_val+self._step_val+1e-7)//self._step_val):
            for i in xrange(int((self._end_val-self._start_val+self._step_val+1e-7)//self._step_val)):
                val = self._start_val + i*self._step_val
                if not self.filter is None and not apply_indexed_rule(self, self.filter, self._parent(), val):
                    continue
                if not self.validate is None and not apply_indexed_rule(self, self.validate, self._parent(), val):
                    continue
                yield val

    def data(self):
        """The underlying set data."""
        return set(self)

    def first(self):
        return self._bounds[0]

    def last(self):
        return self._bounds[1]

    def member(self, key):
        if key >= 1:
            if key > self._len:
                raise IndexError, "Cannot index a RangeSet past the last element"
            return self._start_val + (key-1)*self._step_val
        elif key < 0:
            if self._len+key < 0:
                raise IndexError, "Cannot index a RangeSet past the first element"
            return self._start_val + (self._len+key)*self._step_val
        else:
            raise IndexError, "Valid index values for sets are 1 .. len(set) or -1 .. -len(set)"

    def __eq__(self, other):
        """ Equality comparison """
        if other is None:
            return False
        tmp = self._set_repn(other)
        if self.dimen != other.dimen:
            return False
        ctr = 0
        for i in self:
            if not i in other:
                return False
            ctr += 1
        return ctr == len(other)

    def _set_contains(self, element):
        # As the test for type -- or conversion to float -- is
        # expensive, we will work on the assumption that folks will "do
        # the right thing"... and let system exceptions handle the case
        # where they do not.
        
        #if not type(element) in [int,float]:
        #    return False
        try:
            x = element - self._start_val
            if x % self._step_val != 0:
                # If we are doing floating-point arithmetic, there is a
                # chance that we are seeing roundoff error...
                if math.fabs((x + 1e-7) % self._step_val) > 2e-7:
                    return False
            if element < self._bounds[0] or element > self._bounds[1]:
                return False
        except:
            return False
        if self.filter is not None and not self.filter(element):
            return False
        if self.validate is not None and not self.validate(self, element):
            return False
        return True




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2010 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

__all__ = ['linearize_model_expressions', 'is_linear_expression_constant']

# a prototype utility to translate from a "normal" model instance containing
# constraints and objectives with expresssions for their bodies, into a
# more compact linearized verson of the problem. eliminates the need for
# canonical expressions and the associated overhead (memory and speed).

# NOTE: Currently only works for constraints - leaving the expressions alone,
#       mainly because PH doesn't modify those and I don't want to deal with
#       that aspect for the moment. In particular, this only works for
#       immutable parameters due to the reliance on the canonical expression
#       generator to extract the linear and constant terms.

from coopr.pyomo import *
from coopr.pyomo.expr import generate_canonical_repn, canonical_is_linear, canonical_is_constant

def linearize_model_expressions(instance):
    var_id_map = {}

    # TBD: Should we really be doing this for all components, and not just active ones?
    for constraint_name, constraint in instance.active_components()[Constraint].iteritems():

        for index, constraint_data in constraint._data.iteritems():

            if constraint_data.repn != None:
                canonical_encoding = constraint_data.repn
            else:
                canonical_encoding = generate_canonical_repn(constraint_data.body, var_id_map)

            variable_map = canonical_encoding[-1]

            # we obviously can't linearize an expression if it has higher-order terms!
            if canonical_is_linear(canonical_encoding) or canonical_is_constant(canonical_encoding):

                constant_term = 0.0
                linear_terms = [] # a list of coefficient, _VarData pairs.

                if canonical_encoding.has_key(0):
                    constant_term = canonical_encoding[0][None]

                for var_hash, var_coefficient in canonical_encoding[1].iteritems():
                    var_value = variable_map[var_hash]
                    linear_terms.append((var_coefficient, var_value))

                # eliminate the expression tree - we don't need it any longer.
                # ditto the canonical representation.
                constraint_data.body = None
                constraint_data.lin_body = [constant_term, linear_terms]
                if constraint_data.repn is not None:
                    constraint_data.repn = None

def is_linear_expression_constant(lin_body):
    """Return True if the linear expression is a constant expression, due either to a lack of variables or fixed variables"""
    return (len(lin_body[1]) == 0)



from coopr.pyomo import *
from coopr.pyomo.transform import Transformation

__all__ = ["LinearTransformation"]

class LinearTransformation(Transformation):
    """ Base class for all linear model transformations. """

    def __init__(self, **kwds):
        kwds["name"] = kwds.get("name", "linear_transform")
        super(LinearTransformation, self).__init__(**kwds)



#
# Test the Pyomo command-line interface
#

import unittest
import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep
scriptdir = dirname(dirname(dirname(dirname(dirname(dirname(abspath(__file__)))))))+os.sep

import coopr.pyomo
import pyutilib.subprocess
import pyutilib.th as unittest



class Test(unittest.TestCase):

    def run_convert2nl(self, cmd, file=None):
        os.chdir(currdir)
        return pyutilib.subprocess.run('pyomo2nl --symbolic-solver-labels '+cmd, outfile=file)

    def run_convert2lp(self, cmd, file=None):
        os.chdir(currdir)
        return pyutilib.subprocess.run('pyomo2lp --symbolic-solver-labels '+cmd, outfile=file)

    def test_affine_lp(self):
        """Test examples/pyomo/piecewise/affine.py"""
        self.run_convert2lp(os.path.join(scriptdir,'examples','pyomo','piecewise','affine.py'), file=currdir+'affine.out')
        self.assertFileEqualsBaseline(currdir+'unknown.lp', currdir+'affine.lp')
        os.remove(currdir+'affine.out')

    def test_affine_nl(self):
        """Test examples/pyomo/piecewise/affine.py"""
        self.run_convert2nl(os.path.join(scriptdir,'examples','pyomo','piecewise','affine.py'), file=currdir+'affine.out')
        self.assertFileEqualsBaseline(currdir+'unknown.nl', currdir+'affine.nl')
        os.remove(currdir+'unknown.row')
        os.remove(currdir+'unknown.col')
        os.remove(currdir+'affine.out')

    def test_convex_lp(self):
        """Test examples/pyomo/piecewise/convex.py"""
        self.run_convert2lp(os.path.join(scriptdir,'examples','pyomo','piecewise','convex.py'), file=currdir+'convex.out')
        self.assertFileEqualsBaseline(currdir+'unknown.lp', currdir+'convex.lp')
        os.remove(currdir+'convex.out')

    def test_convex_nl(self):
        """Test examples/pyomo/piecewise/convex.py"""
        self.run_convert2nl(os.path.join(scriptdir,'examples','pyomo','piecewise','convex.py'), file=currdir+'convex.out')
        self.assertFileEqualsBaseline(currdir+'unknown.nl', currdir+'convex.nl')
        os.remove(currdir+'unknown.row')
        os.remove(currdir+'unknown.col')
        os.remove(currdir+'convex.out')

    def test_indexed_lp(self):
        """Test examples/pyomo/piecewise/indexed.py"""
        self.run_convert2lp(os.path.join(scriptdir,'examples','pyomo','piecewise','indexed.py'), file=currdir+'indexed.out')
        self.assertFileEqualsBaseline(currdir+'unknown.lp', currdir+'indexed.lp')
        os.remove(currdir+'indexed.out')

    def test_indexed_nl(self):
        """Test examples/pyomo/piecewise/indexed.py"""
        self.run_convert2nl(os.path.join(scriptdir,'examples','pyomo','piecewise','indexed.py'), file=currdir+'indexed.out')
        self.assertFileEqualsBaseline(currdir+'unknown.nl', currdir+'indexed.nl')
        os.remove(currdir+'unknown.row')
        os.remove(currdir+'unknown.col')
        os.remove(currdir+'indexed.out')

    def test_indexed_nonlinear_nl(self):
        """Test examples/pyomo/piecewise/indexed_nonlinear.py"""
        self.run_convert2nl(os.path.join(scriptdir,'examples','pyomo','piecewise','indexed_nonlinear.py'), file=currdir+'indexed_nonlinear.out')
        self.assertFileEqualsBaseline(currdir+'unknown.nl', currdir+'indexed_nonlinear.nl')
        os.remove(currdir+'unknown.row')
        os.remove(currdir+'unknown.col')
        os.remove(currdir+'indexed_nonlinear.out')


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2010 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

__all__ = ['label_from_name', 'nl_label_from_name']

import string


# This module provides some basic functionality for generating labels
# from pyomo names, which often contain characters such as "[" and "]"
# (e.g., in my_var[1]).  These characters generally cause issues with
# optimization input file formats, e.g., CPLEX LP files.  The purpose of
# this module is to provide a simple remap function, that will satisfy
# broadly problematic symbols. if solver-specific remaps are required,
# they should be handled either in the corresponding solver plugin.

# NOTE: Simple single-character substitutions should be handled by adding
#       to the translation table constructed below - first argument is the
#       "from" characters, second argument is the "to" characters.
simple_translation_table = string.maketrans("[]{} -#$%&*+.,/;<=>?@^!~'",
                                            "()()_____________________")

nl_translation_table = string.maketrans("{} -#$%&*+./;<=>?@^!~'",
                                            "()____________________")

def label_from_name(name):

    global simple_translation_table

    if name is None:
        raise RuntimeError(
            "Illegal name=None supplied to label_from_name function")

    return string.translate(name, simple_translation_table)

def nl_label_from_name(name):
    global nl_translation_table
    
    if name is None:
        raise RuntimeError("Illegal name=None supplied to nl_label_from_name function")
    
    return string.translate(name, nl_translation_table)
